package dareka.common;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;

/**
 * [nl] t@C֌W̃[eBeBNXB
 * {IɗO͑SăNXŏ̂ŁA߂lŐۂ𔻒f邱ƁB
 */
public class FileUtil {
	private FileUtil() {} // avoid instantiation
	private static final int BUF_SIZE = 64 * 1024;
	
	/**
	 * t@CԂŃRs[BRs[Ƀt@C݂ꍇ͏㏑B
	 * G[Rs[̃t@C͍폜B
	 * @param src Rs[̃t@C
	 * @param dst Rs[̃t@C
	 * @return Rs[oCgB0oCgL蓾BG[-1
	 */
	public static long copy(File src, File dst) {
		long len = -1;
		if (src != null && dst != null) {
			FileInputStream  in  = null;
			FileOutputStream out = null;
			try {
				// Rs[ƃRs[悪ȂRs[Ȃ
				if (!src.getCanonicalPath().equals(dst.getCanonicalPath())) {
					in = new FileInputStream(src);
					FileChannel srcCh = in.getChannel();
					out = new FileOutputStream(dst);
					FileChannel dstCh = out.getChannel();
					len = srcCh.transferTo(0, srcCh.size(), dstCh);
				}
			} catch (IOException e) {
				Logger.debugWithThread(e);
			} finally {
				CloseUtil.close(in);
				CloseUtil.close(out);
				if (len < 0) dst.delete();
			}
		}
		return len;
	}
	
	/**
	 * Xg[t@CɃRs[BRs[Ƀt@C݂ꍇ͏㏑B
	 * G[Rs[̃t@C͍폜B
	 * @param src Rs[̃Xg[
	 * @param dst Rs[̃t@C
	 * @return Rs[oCgB0oCgL蓾BG[-1
	 */
	public static int copy(InputStream src, File dst) {
		int len = -1;
		if (src != null && dst != null) {
			FileOutputStream out = null;
			try {
				out = new FileOutputStream(dst);
				len = 0;
				byte[] buf = new byte[BUF_SIZE];
				for (int n; (n = src.read(buf)) != -1; len += n)
					out.write(buf, 0, n);
			} catch (IOException e) {
				Logger.debugWithThread(e);
				len = -1;
			} finally {
				CloseUtil.close(out);
				if (len < 0) dst.delete();
			}
		}
		return len;
	}
	
	/**
	 * t@CXg[ɃRs[B
	 * @param src Rs[̃t@C
	 * @param dst Rs[̃Xg[
	 * @return Rs[oCgB0oCgL蓾BG[-1
	 */
	public static int copy(File src, OutputStream dst) {
		int len = -1;
		if (src != null && dst != null) {
			FileInputStream in = null;
			try {
				in = new FileInputStream(src);
				len = 0;
				byte[] buf = new byte[BUF_SIZE];
				for (int n; (n = in.read(buf)) != -1; len += n)
					dst.write(buf, 0, n);
			} catch (IOException e) {
				Logger.debugWithThread(e);
				len = -1;
			} finally {
				CloseUtil.close(in);
			}
		}
		return len;
	}
	
	/**
	 * obt@t@CɃRs[BRs[Ƀt@C݂ꍇ͏㏑B
	 * G[Rs[̃t@C͍폜B
	 * @param src Rs[̃obt@
	 * @param dst Rs[̃t@C
	 * @return Rs[oCgB0oCgL蓾BG[-1
	 */
	public static int copy(ByteBuffer src, File dst) {
		int len = -1;
		if (src != null && dst != null) {
			FileOutputStream out = null;
			try {
				out = new FileOutputStream(dst);
				FileChannel ch = out.getChannel();
				len = ch.write(src);
				ch.close();
			} catch (IOException e) {
				Logger.debugWithThread(e);
			} finally {
				CloseUtil.close(out);
				if (len < 0) dst.delete();
			}
		}
		return len;
	}
	
	/**
	 * t@Cobt@ɃRs[BRs[̃obt@̓Rs[JnʒuɃ}[NB
	 * G[Rs[̃obt@̓}[NJnʒu܂ŃZbgB
	 * @param src Rs[̃t@C
	 * @param dst Rs[̃obt@
	 * @return Rs[oCgB0oCgL蓾BG[-1
	 */
	public static int copy(File src, ByteBuffer dst) {
		int len = -1;
		if (src != null && dst != null) {
			dst.mark();
			FileInputStream in = null;
			try {
				in = new FileInputStream(src);
				FileChannel ch = in.getChannel();
				len = ch.read(dst);
				ch.close();
			} catch (IOException e) {
				Logger.debugWithThread(e);
			} finally {
				CloseUtil.close(in);
				if (len < 0) dst.reset();
			}
		}
		return len;
	}
	
	/**
	 * Xg[Xg[ɃRs[B
	 * @param src Rs[̃Xg[
	 * @param dst Rs[̃Xg[
	 * @return Rs[oCgB0oCgL蓾BG[-1
	 */
	public static int copy(InputStream src, OutputStream dst) {
		int len = -1;
		if (src != null && dst != null) {
			try {
				len = 0;
				byte[] buf = new byte[BUF_SIZE];
				for (int n; (n = src.read(buf)) != -1; len += n)
					dst.write(buf, 0, n);
			} catch (IOException e) {
				Logger.debugWithThread(e);
				len = -1;
			}
		}
		return len;
	}
	
	/**
	 * t@C̕Zbg𔻕ʂ{@link java.io.InputStreamReader}𐶐ĕԂB
	 * Zbg𔻕ʂ邽߂ɁA{ꕶ܂񂾊Jnw肷KvB
	 * <br>
	 * ȉ̂ꂩ̏ꍇ̓ftHgZbg{@link java.io.InputStreamReader}𐶐B
	 * <ul>
	 * <li>JnŎn܂ĂȂ
	 * <li>Jnɓ{ꕶ܂܂ĂȂ
	 * <li>JnnullA͋󕶎
	 * <li>t@CTCY
	 * </ul>
	 * @param file Ώۂ̃t@C
	 * @param startline Zbg̃eXgɎgJn
	 * @return {@link java.io.InputStreamReader InputStreamReader}̃CX^X
	 * @throws IOException t@C̓o͂Ŗ肪
	 */
	public static InputStreamReader getInputStreamReader(
			File file, String startline) throws IOException {
		if (file == null) {
			throw new IllegalArgumentException("file must not be null");
		}
		BufferedInputStream in = new BufferedInputStream(new FileInputStream(file));
		if (startline != null) {
			try {
				Info info = getInfo(file, null, startline, in);
				return new InputStreamReader(in, info.charset);
			} catch (IOException e) {
				CloseUtil.close(in);
				throw e;
			}
		}
		return new InputStreamReader(in);
	}
	
	/**
	 * t@C̕Zbg𔻕ʂ{@link java.io.OutputStreamWriter}𐶐ĕԂB
	 * Zbg̔ʏ{@link #getInputStreamReader(File, String)}QƂ̂ƁB
	 * t@C݂ꍇappend[hŃt@CI[v܂B
	 * t@C݂Ȃꍇ̓ftHgZbgŃt@C쐬܂B
	 * 
	 * @param file Ώۂ̃t@C
	 * @param startline Zbg̃eXgɎgJn
	 * @return {@link java.io.OutputStreamWriter OutputStreamWriter}̃CX^X
	 * @throws IOException  t@C̓o͂Ŗ肪
	 * @see #getInputStreamReader(File, String)
	 */
	public static OutputStreamWriter getOutputStreamWriter(
			File file, String startline) throws IOException {
		if (file == null) {
			throw new IllegalArgumentException("file must not be null");
		}
		if (!file.exists() || startline == null) {
			return new OutputStreamWriter(new FileOutputStream(file));
		}
		Info info;
		BufferedInputStream in = new BufferedInputStream(new FileInputStream(file));
		try {
			info = getInfo(file, null, startline, in);
		} finally {
			CloseUtil.close(in);
		}
		return new OutputStreamWriter(
				new FileOutputStream(file, true), info.charset);
	}
	
	// TODO 낢됮
	public static class Info {
		public long lastModified;
		public String startline, line_separator;
		public Charset charset;
		
		public Info(File file) {
			this(file, null);
		}
		public Info(File file, String startline) {
			this.lastModified = file.lastModified();
			this.startline = startline;
		}
	}
	
	public static Info getInfo(File file, Info info, String startline,
			BufferedInputStream in) throws IOException {
		info = getInfo(file, info, startline);
		if (in == null) return info;
		
		long length = file.length();
		byte[] testBuf = new byte[length > 256L ? 256 : (int) length];
		in.mark(testBuf.length);
		in.read(testBuf);
		in.reset();
		String s = getCharsetNameFromBOM(testBuf, in);
		if (s != null) {
			info.charset = Charset.forName(s);
		} else if (info.startline != null) {
			// 14oCgŔ
			int testSize = info.startline.length() * 4;
			if (testSize > testBuf.length) {
				testSize = testBuf.length;
			}
			for (Charset c : SUPPORTED_CHARSET) {
				s = new String(testBuf, 0, testSize, c.name()); // JDK1.1
				if (s.startsWith(info.startline)) {
					info.charset = c; break;
				}
			}
		}
		if (info.charset == null) {
			info.charset = Charset.defaultCharset();
		}
		info.line_separator = guessLineSeparator(testBuf);
		
		Logger.debugWithThread(
				file.getPath() + ": charset=" + info.charset +
				getLineSeparatorSymbol(info.line_separator));
		return info;
	}
	
	public static Info getInfo(File file, Info info, String startline) {
		if (file == null) {
			throw new IllegalArgumentException("file must not be null");
		}
		if (info == null) {
			info = new Info(file, startline);
		} else {
			long lastModified = file.lastModified();
			if (info.lastModified != lastModified) {
				info.lastModified = lastModified;
				info.charset = null;
				info.line_separator = null;
			}
		}
		return info;
	}
	
	private static final Charset[] SUPPORTED_CHARSET = {
			Charset.forName("MS932"),
			Charset.forName("UTF-8"),
			Charset.forName("UTF-16BE"),
			Charset.forName("UTF-16LE"),
			Charset.forName("EUC-JP")
	};
	
	private static String getCharsetNameFromBOM(
			byte[] buf, BufferedInputStream in) throws IOException {
		if (buf.length < 3) {
			return null;
		}
		if (buf[0] == (byte)0xEF && buf[1] == (byte)0xBB && buf[2] == (byte)0xBF) {
			if (in != null) in.skip(3);
			return "UTF-8";
		} else if (buf[0] == (byte)0xFE && buf[1] == (byte)0xFF) {
			if (in != null) in.skip(2);
			return "UTF-16BE";
		} else if (buf[0] == (byte)0xFF && buf[1] == (byte)0xFE) {
			if (in != null) in.skip(2);
			return "UTF-16LE";
		}
		return null;
	}
	
	public static String guessLineSeparator(byte[] testBuf) {
		int cr = 0, lf = 0;
		for (int i = 0; i < testBuf.length; i++) {
			if (testBuf[i] == 0x0a) {
				lf++;
			} else if (testBuf[i] == 0x0d) {
				cr++;
			}
		}
		if (cr > 0 && lf > 0 && Math.abs(cr - lf) < 2) {
			return "\r\n";
		} else if (cr < lf) {
			return "\n";
		} else if (cr > lf) {
			return "\r";
		}
		return System.getProperty("line.separator");
	}
	
	private static String getLineSeparatorSymbol(String ls) {
		if ("\r\n".equals(ls)) {
			return " CRLF";
		} else if ("\n".equals(ls)) {
			return " LF";
		} else if ("\r".equals(ls)) {
			return " CR";
		}
		return "";
	}
}
