package dareka.processor;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.Logger;
import dareka.extensions.SystemEventListener;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.URLResource;
import dareka.processor.impl.NLEventSource;
import dareka.processor.impl.NLShared;

/**
 * [nl] URLResource ƃL[֘AtăɕێNXB
 * 
 * yɎMeLbV邽߂̎i񋟂B
 * @ɂĂ {@link dareka.processor.impl.GetThumbInfoProcessor} QƂ̂ƁB
 */
public class URLResourceCache {
	
	private class LRUMap extends LinkedHashMap<String, URLResource> {
		private static final long serialVersionUID = 1L;
		public LRUMap() {
			super(maxEntries > 12 ? (int)(maxEntries * 1.4) : 16, 0.75f, true);
		}
		protected boolean removeEldestEntry(Map.Entry<String, URLResource> eldest) {
			boolean remove = size() > updateMaxEntries();
			if (remove) {
				debugMes("url cache removed", eldest.getKey(), eldest.getValue());
			}
			return remove;
		}
	}
	
	private static final int MAX_ENTRIES_DEFAULT = 10;
	private static final long EXPIRE_TIME_DEFAULT = 10 * 60 * 1000; // 10
	private static final NLShared NLSHARED = NLShared.getInstance();
	private int maxEntries;
	private long expireTime;
	private String maxEntriesConfigKey;
	private String expireTimeConfigKey;
	private Map<String, URLResource> resources;
	
	
	/**
	 * ݒŒ URLResourceCache 쐬B
	 * 
	 * @param maxEntries LbVɕێő吔
	 * @param expireTime Ăj܂ł̎(Pʂ̓~b)
	 * @see dareka.processor.URLResource
	 * @see #expires()
	 */
	public URLResourceCache(int maxEntries, long expireTime) {
		init(maxEntries, expireTime);
	}
	
	/**
	 * ݒύX\ URLResourceCache 쐬B
	 * 
	 * @param maxEntriesConfigKey maxEntries ̒l擾VXevpeB
	 * @param expireTimeConfigKey expireTime ̒l擾VXevpeB(Pʂ͕b)
	 */
	public URLResourceCache(String maxEntriesConfigKey, String expireTimeConfigKey) {
		this.maxEntriesConfigKey = maxEntriesConfigKey;
		this.expireTimeConfigKey = expireTimeConfigKey;
		init(updateMaxEntries(), updateExpireTime());
	}
	
	private void init(int maxEntries, long expireTime) {
		this.maxEntries = maxEntries < 0 ? MAX_ENTRIES_DEFAULT : maxEntries;
		this.expireTime = expireTime < 0 ? EXPIRE_TIME_DEFAULT : expireTime;
		resources = Collections.synchronizedMap(new LRUMap());
	}
	
	private int updateMaxEntries() {
		if (maxEntriesConfigKey != null) {
			maxEntries = Integer.getInteger(maxEntriesConfigKey, -1);
			if (maxEntries < 0) maxEntries = MAX_ENTRIES_DEFAULT;
		}
		return maxEntries;
	}
	
	private long updateExpireTime() {
		if (expireTimeConfigKey != null) {
			expireTime = Long.getLong(expireTimeConfigKey, -1) * 1000;
			if (expireTime < 0) expireTime = EXPIRE_TIME_DEFAULT;
		}
		return expireTime;
	}
	
	/**
	 * LbV URLResource 擾B
	 * 
	 * @param key 擾 URLResource ɑΉL[
	 * @return 擾 URLResourceALbVɑ݂Ȃ null
	 */
	public URLResource get(String key) {
		URLResource r = resources.get(key);
		if (r != null) {
			synchronized (r) { // M܂ŃubN
				if (remainTime(r) == 0) {
					debugMes("url cache expired", key, r);
					resources.remove(key);
					r = null;
				} else {
					debugMes("using url cache", key, r);
				}
			}
		}
		return r;
	}
	
	/**
	 * LbV URLResource 擾B
	 * LbVɑ݂ȂΎw URL MĕԂ(GET̂)B
	 * 
	 * @param key 擾 URLResource ɑΉL[
	 * @param url LbVɑ݂ȂꍇɎM URL
	 * @return 擾 URLResourceAMłȂ null
	 */
	public URLResource cacheAndGet(String key, String url) {
		return cacheAndGet(key, url, null, null);
	}
	
	/**
	 * LbV URLResource 擾B
	 * LbVɑ݂ȂΎw胊NGXgwb_ɎMĕԂ(POSTΉ)B
	 * 
	 * @param key 擾 URLResource ɑΉL[
	 * @param receiverIn POST ̏ꍇ̓̓Xg[
	 * @param requestHeader NGXgwb_
	 * @return 擾 URLResourceAMłȂ null
	 */
	public URLResource cacheAndGet(String key,
			InputStream receiverIn, HttpRequestHeader requestHeader) {
		return cacheAndGet(key, null, receiverIn, requestHeader);
	}
	
	// URL̃p[^Ƀ}b`p^[(Ď擾p)
	private static final Pattern PARAMS_REGET_PATTERN = Pattern.compile(
			"^(https?://.+)[\\?&]\\d+$");
	
	/**
	 * LbV Rewriter  Resource 擾B
	 * LbVɑ݂ȂΎw胊NGXgwb_ɎMĕԂ(POSTΉ)B
	 * URL ̃p[^ŏIꍇALbVjčĎ擾B
	 * 
	 * @param key 擾 URLResource ɑΉL[
	 * @param receiverIn POST ̏ꍇ̓̓Xg[
	 * @param requestHeader NGXgwb_
	 * @return Rewriter  ResourceAMłȂ null
	 * @throws IOException o̓G[ꍇ
	 */
	public Resource cacheAndRewrite(String key,
			InputStream receiverIn, HttpRequestHeader requestHeader) throws IOException {
		String uri = requestHeader.getURI();
		Matcher m = PARAMS_REGET_PATTERN.matcher(uri);
		if (m.matches()) {
			uri = m.group(1);
			remove(key); // LbVj
		}
		URLResource r = cacheAndGet(key, uri, null, requestHeader);
		if (r != null) {
			HttpResponseHeader rh = r.getResponseHeader(null, null);
			if (rh.getStatusCode() != 200)
				return r; // 200ȊOȂRewriter̂܂ܕԂ
			byte[] original = r.getResponseBody();
			if (original != null && original.length > 0) {
				// Rewriterɂ̓p[^݂œn
				byte[] replaced = dareka.Main.getRewriterProcessor().contentRewriter(
						uri, original, requestHeader, rh);
				if (original != replaced) {
					return StringResource.getRawResource(rh, replaced);
				}
			}
			return r;
		}
		return Resource.get(Resource.Type.URL, uri);
	}
	
	private URLResource cacheAndGet(String key, String url,
			InputStream receiverIn, HttpRequestHeader requestHeader) {
		URLResource r = get(key);
		if (r == null) {
			if (url == null && requestHeader != null) {
				url = requestHeader.getURI();
			}
			if (url == null) return null;
			try {
				r = new URLResource(url); // bNIuWFNg˂
				synchronized (r) {
					// Ƃ肠}bvɓo^đdMȂ悤ɂ
					resources.put(key, r);
					
					doCache(key, url, receiverIn, requestHeader, r);
				}
			} catch (IOException e) {
				resources.remove(key);
				r = null;
				Logger.error(e);
			}
		}
		return r;
	}
	
	private void doCache(String key, String url,
			InputStream receiverIn, HttpRequestHeader requestHeader,
			URLResource r) throws IOException {
		r.setFollowRedirects(true);
		HttpResponseHeader responseHeader = r.getResponseHeader(
				receiverIn, requestHeader);
		if (responseHeader != null) {
			String date = responseHeader.getMessageHeader(HttpHeader.DATE);
			if (date == null) {
				// Expirê߂Datewb_t
				date = HttpHeader.getDateString(System.currentTimeMillis());
				responseHeader.setMessageHeader(HttpHeader.DATE, date);
			}
			r.getResponseBody();
			
			debugMes("url cached", key, r);
			
			if (NLSHARED.countSystemEventListeners() > 0) {
				NLSHARED.notifySystemEvent(
						SystemEventListener.URL_MEMCACHED,
						new NLEventSource(url, requestHeader, r), false);
			}
		} else {
			resources.remove(key); // wb_MłȂΔj
		}
	}
	
	/**
	 * URLResource LbVĂ邩H
	 * 
	 * @param key URLResource ɑΉL[
	 * @return key ɑΉ URLResource LbVĂ true
     * @since NicoCache_nl+110706mod
	 */
	public boolean isCached(String key) {
		return resources.containsKey(key);
	}
	
	/**
	 * LbV URLResource 폜B
	 * 
	 * @param key 폜 URLResource ɑΉL[
	 */
	public void remove(String key) {
		debugMes("url cache removed", key, null);
		resources.remove(key);
	}
	
	/**
	 * LbVɕێĂURLResource̐ԂB
	 * 
	 * @return ێĂURLResource̐
	 */
	public int size() {
		return resources.size();
	}
	
	/**
	 * LbVNAB
	 */
	public void clear() {
		resources.clear();
	}
	
	/**
	 * LbV̑SĂ URLResource ɂāAexpireTime ɊÂČÂ
	 * jBȂAexpireTime ̓X|XĂ̌oߎԂłA
	 * LbVĂ̌oߎԂł͖̂Œӂ邱ƁB
	 * 
	 */
	public synchronized void expires() {
		try {
			Set<Map.Entry<String, URLResource>> s = resources.entrySet();
			for (Map.Entry<String, URLResource> e : s) {
				if (remainTime(e.getValue()) == 0) {
					debugMes("url cache expired", e.getKey(), e.getValue());
					s.remove(e);
				}
			}
		} catch (ConcurrentModificationException e) {
			// s삪ꍇ͎Ɏz
		}
	}
	
	private long remainTime(URLResource resource) {
		long et = updateExpireTime();
		if (et > 0 && resource != null) {
			// TODO LbVwb_l
			HttpResponseHeader rh = getResponseHeader(resource);
			if (rh != null) {
				long tm = -1;
				String date = rh.getMessageHeader(HttpHeader.DATE);
				if (date != null) {
					tm = HttpHeader.parseDateString(date);
				}
				if (rh.getStatusCode() != 200) {
					et = 30000; // Xe[^X200ȊOȂ30bŔj
				}
				long remain = et + tm - System.currentTimeMillis();
				if (remain > 0) return remain;
			}
		}
		return 0;
	}
	
	private HttpResponseHeader getResponseHeader(URLResource r) {
		if (r != null) {
			try {
				return r.getResponseHeader(null, null);
			} catch (IOException e) {}
		}
		return null;
	}
	
	private void debugMes(String mes, String key, URLResource r) {
		if (Boolean.getBoolean("dareka.debug")) {
			long remain = remainTime(r);
			Logger.debugWithThread(mes + ": key=" + key + " remain=" + remain);
		}
	}
	
}
