package dareka.processor.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.NLConfig;
import dareka.common.FileUtil;
import dareka.common.Logger;
import dareka.common.TextUtil;
import dareka.extensions.SystemEventListener;


// bNpStaticIuWFNg
class LockObj {
    static Object FileMoveLock = new Object();
    static Object FileStoreLock = new Object();
}

/**
 *
 * Abstraction of cache handling.
 *
 * Naming specification:
 *
 * A cache is identified by cacheId which consists of letters without '_'
 * (except nltmp).
 *
 * A cache may have a description which consists of letters being available for
 * filename.
 *
 * Mapping of cache names and filename: cacheId + ('_' + description)? + postfix
 *
 */
// TODO Too fragile! Have to be managed the state... 
public class Cache extends CacheManager {
    public static final String FLV = ".flv";
    public static final String SWF = ".swf";
    public static final String MP4 = ".mp4";

    private String cacheId;
    private String postfix, desc;
    private File cacheFileNew; // [nl] VKɍ쐬ꍇ̃LbVt@C
    private File tmpFile;

    public static void init() {
        CacheManager.init();
    }

	// [nl] tH_Ԉړ
	public boolean moveTo(String dir) {
		if (dirList.contains(dir)) {
			File cacheFile = getCacheFile();
			File toFile = new File(dir2File.get(dir), cacheFile.getName());
			return moveCache(cacheFile, toFile);
		}
		return false;
	}
	
	// [nl] LbVt@CɈړ
	private boolean moveCache(File src, File dst) {
		if (src != null && dst != null &&
				src.exists() && !dst.exists() && !src.equals(dst)) {
			// ܂l[Ă݂
			id2FileLock.writeLock().lock();
			try {
				if (src.renameTo(dst)) {
					id2File_put(cacheId, dst);
					return true;
				}
			} finally {
				id2FileLock.writeLock().unlock();
			}
			// ʖڂȂRs[Ă݂B
			// 
			// ڃRs[init()ĂяoAPIsĔsꂽꍇ
			// Rs[r̂̂FĕsN\̂ŁA
			// ܂ꎞt@CɃRs[Ă߂ăl[B
			// ܂ARs[̏ꍇI/O𒼗񉻂ċԂhB
			// 
			File tmp = new File(dst.getPath().concat(".tmp"));
			synchronized (LockObj.FileMoveLock) {
				if (tmp.exists() || FileUtil.copy(src, tmp) == -1)
					return false;
				// Rs[ĂԂsrcύXꍇ邩
				if (!src.exists()) {
					if ((src = id2File_get(cacheId)) == null || !src.exists()) {
						tmp.delete();
						return false;
					}
					dst = new File(dst.getParentFile(), src.getName());
				}
			}
			id2FileLock.writeLock().lock();
			try {
				if (tmp.renameTo(dst)) {
					src.delete();
					id2File_put(cacheId, dst);
					return true;
				} else {
					tmp.delete(); // s(܂L蓾ȂƎvǁc)
				}
			} finally {
				id2FileLock.writeLock().unlock();
			}
		}
		return false; 
	}
	
	// [nl] LbVt@CwpXɃRs[
	public boolean copyTo(String dstPath) {
		return copyTo(new File(dstPath));
	}
	public boolean copyTo(File dstFile) {
		File cacheFile = id2File_get(cacheId);
		if (cacheFile != null && cacheFile.exists()) {
			return FileUtil.copy(cacheFile, dstFile) != -1;
		}
		return false;
	}
	
	// [nl] wpXLbVt@CɃRs[
	public boolean copyFrom(String srcPath) {
		return copyFrom(new File(srcPath));
	}
	public boolean copyFrom(File srcFile) {
		File cacheFile = getCacheFile();
		// Ue|t@CɃRs[Đ΃LbVɒu
		File tmp = new File(cacheFile.getPath().concat(".tmp"));
		if (!tmp.exists() && FileUtil.copy(srcFile, tmp) != -1) {
			id2FileLock.writeLock().lock();
			try {
				cacheFile.delete();
				if (tmp.renameTo(cacheFile)) {
					id2File_put(cacheId, cacheFile);
					return true;
				}
			} finally {
				id2FileLock.writeLock().unlock();
			}
		}
		return false;
	}
	
	/**
	 *  [nl] LbV菜BExtension jQ邱ƂB
	 * @param id LbV ID
	 * @return LbV菜 true
	 */
	public static boolean remove(String id) {
		File cacheFile = id2File_get(id);
		if (cacheFile == null || !cacheFile.exists()) {
			return false;
		}
		NLEventSource eventSource = null;
		if (NLShared.INSTANCE.countSystemEventListeners() > 0) {
			eventSource = new NLEventSource(null, null, new Cache(id));
			int result = NLShared.INSTANCE.notifySystemEvent(
					SystemEventListener.CACHE_REMOVING, eventSource, true);
			if (result != SystemEventListener.RESULT_OK) {
				return false;
			}
		}
		if (cacheFile.delete()) {
			id2File_put(id, null);
			if (eventSource != null) {
				NLShared.INSTANCE.notifySystemEvent(
						SystemEventListener.CACHE_REMOVED, eventSource, false);
			}
			return true;
		}
		return false;
	}
	
	// [nl] ꎞt@C菜
	public static boolean removeTmp(String id) {
		File tmp = id2Tmp.get(id);
		if (tmp != null && getDLFlag(id) == false) {
			if (tmp.exists() && tmp.delete()) {
				id2Tmp.remove(id);
				//[xmlfix]
				removeXmlCache(TYPE_TEMPLIST, null);
				return true;
			}
		}
		return false;
	}
	
	// [nl] wIDɊ֘Aȉ̃t@CSĎ菜
	// ʏLbVAGRm~[LbVÄꎞt@C
	public static boolean removeAll(String id) {
		return removeAll(id, true);
	}
	public static boolean removeAll(String id, boolean rmDL) {
		boolean removed = false;
		String[] s = parseSmid(id);
		if (s != null) {
			if(rmDL) removed |= removeDL(s[0], s[1]);
			removed |= remove(s[0]) | remove(s[1]) |
					removeTmp(s[0]) | removeTmp(s[1]);
		}
		return removed;
	}
	
	// [nl] _E[h̃LbV폜\񂷂
	private static boolean removeDL(String smid, String lowId) {
		if (Cache.getDLFlag(smid) || Cache.getDLFlag(lowId)) {
			NLShared.INSTANCE.addDeleteSet(smid);
			//XgɒǉĂԂɃ_E[hIĂ폜
			//DLtO~낷ʒuύX疳Ăǂ͂c
			if (!Cache.getDLFlag(smid) && !Cache.getDLFlag(lowId)) {
				NLShared.INSTANCE.removeDeleteSet(smid);
				removeAll(smid, false);
			}
			return true;
		}
		return false;
	}
	
	// [nl] URL̎ʎqgqZbg
	protected void setPostfix(String vType) {
		switch (vType.charAt(0)) {
		case 'v':
			postfix = FLV; break;
		case 's':
			postfix = SWF; break;
		case 'm':
			postfix = MP4; break;
		case '.':
			postfix = vType; break;
		}
	}
	
	public String getPostfix() {
		return postfix;
	}

    public static void cleanup() {
        // [nl] ꎞt@C폜(Ȃ)
    }

    // [nl] RXgN^
    public Cache(String cacheId) {
        this(cacheId, FLV);
    }

    public Cache(String cacheId, String postfix) {
        this.cacheId = cacheId;
        setPostfix(postfix);
/* [nl] r䂪KvȂ̂cacheFile͕ێȂ
        cacheFile = id2File.get(cacheId);
        if (cacheFile == null) {
            cacheFile = new File(cacheDir, cacheId + postfix);
        }*/
    }

    public Cache(String cacheId, String postfix, String desc) {
        this.cacheId = cacheId;
        this.desc = TextUtil.stripZeroWithChars(desc); // [nl]
        setPostfix(postfix);
/* [nl] r䂪KvȂ̂cacheFile͕ێȂ
        cacheFile = id2File.get(cacheId);
        if (cacheFile == null) {
            cacheFile = getDescribedCacheFile(desc);
        }*/
    }

    // [nl] VKLbVt@C̃CX^X쐬
    // RXgN^ō쐬ɎۂɕKvɂȂ鎞܂Œx
    private void setCacheFileNew() {
    	if (desc != null) {
    		cacheFileNew = getDescribedCacheFile(desc);
    	} else {
    		cacheFileNew = new File(cacheDir, cacheId + postfix);
    	}
    }

	// [nl] ^Cg擾iꍇIDj
	public String getTitle() {
		String filename = this.getCacheFileName();
		Matcher m = CACHE_FILE_PATTERN.matcher(filename);
		if (m.find()) {
			if (m.group(2) != null)
				return m.group(2);
			else
				return m.group(1);
		} else {
			return "";
		}
	}
	
	// [nl] t@C̃^CgύX
	public boolean setTitle(String newTitle) {
		desc = newTitle;
		setCacheFileNew();
		File cacheFile = id2File_get(cacheId);
		File newFile = new File(cacheFile.getParentFile(), cacheFileNew.getName());
		return moveCache(cacheFile, newFile);
	}

    private class DirInfo implements Comparable<DirInfo> {
    	public String path;
    	public String name;
    	public int depth;
    	public DirInfo(String path, String name, int depth) {
    		this.path = path;
    		this.name = name;
    		this.depth = depth;
    	}
    	
    	// OE[
    	public int compareTo(DirInfo e) {
    		if (this.name.length() == e.name.length())
    			return this.depth == e.depth ? 0 : (this.depth < e.depth ? 1 : -1);
    		return this.name.length() < e.name.length() ? 1 : -1;
    	}
    }

	// [nl]tH_tB^ƂāAXgAT
	protected String suggestStoreFolder() {
		ArrayList<DirInfo> dirs = new ArrayList<DirInfo>();
		for (String dir : dirList) {
			if (dir.length() == 0 || NLConfig.find("storeFilterIgnore", dir))
				continue;
			String[] parts = dir.split("/");
			int n = parts.length;
			dirs.add(new DirInfo(dir, tidyTitle(parts[n-1].toLowerCase()), n));
		}
		// Ƀ\[gāA "/" Ă
		Collections.sort(dirs);
		
		String test = getCacheFileName().toLowerCase();
		for (DirInfo entry : dirs) {
			String[] patterns = entry.name.split(" +");
			boolean match = true;
			for (String p : patterns) {
				if (p.charAt(0) == '-') {
					// NOT
					if (test.contains(p.substring(1)))
						match = false;
				} else {
					// AND
					if (!test.contains(p))
						match = false;
				}
			}
			if (match)
				return entry.path;
		}
		
		return "";
	}
	
	// [nl] ^CgLx`
	// 		Spppp
	// 		̔pSpXy[X1̔pXy[X
	private static final Pattern spacePattern = Pattern.compile("[ @]+");
	public static String tidyTitle(String oldTitle) {
		char title[] = spacePattern.matcher(oldTitle).replaceAll(" ").trim().toCharArray();
		
		for (int i = 0; i < title.length; i++) {
			char c = title[i];
			if ('O' <= c && c <= 'X')
				c = (char)(c - 'O' + '0');
			else if ('`' <= c && c <= 'y')
				c = (char)(c - '`' + 'A');
			else if ('' <= c && c <= '')
				c = (char)(c - '' + 'a');
			title[i] = c;
		}
		return new String(title);
	}
	
	// [nl] ̎Q(Entity Reference)u
	private static String replaceER(String src) {
		return src.replaceAll("&amp;" , "&")
				  .replaceAll("&lt;"  , "<")
				  .replaceAll("&gt;"  , ">")
				  .replaceAll("&quot;", "\"")
				  .replaceAll("&apos;", "'");
	}

    protected File getDescribedCacheFile(String desc) {
        StringBuffer decodedDesc = new StringBuffer(desc.length());

        Matcher m = NUMBER_CHARACTER_REFERENCE_PATTERN.matcher(desc);

        while (m.find()) {
            m.appendReplacement(decodedDesc,
                    Character.toString((char) Integer.parseInt(m.group(1))));
        }
        m.appendTail(decodedDesc);

        String sanitizedDesc = getSanitizedDescription(decodedDesc.toString());

        // [nl] ^Cg`
        if (Boolean.getBoolean("tidyTitle")) {
        	sanitizedDesc = tidyTitle(sanitizedDesc);
        }

        return new File(cacheDir, cacheId + '_' + sanitizedDesc + postfix);
    }

    public static String getSanitizedDescription(String decodedDesc) {
        String fileNameCharset = System.getProperty("fileNameCharset");

        String narrowedDesc = narrowCharset(decodedDesc, fileNameCharset);

        // ̎QƂXɎ̎QƂP[X̂œu
        narrowedDesc = replaceER(replaceER(narrowedDesc));

        char descArray[] = narrowedDesc.toCharArray();

        for (int i = 0; i < descArray.length; i++) {
        	char c = descArray[i];
            switch (c) {
            case '\\': case '\u00a5': c = ''; break;
            case '/': c = '^';	break;
            case ':':	c = 'F';		break;
            case '*':	c = '';	break;
            case '?':	c = 'H';	break;
            case '"':	c = 'h';	break;
            case '<':	c = '';	break;
            case '>':	c = '';	break;
            case '|':	c = 'b';	break;
            case '\u203e':// c = '~'; break;
            case '\u301c':// c = '`'; break;
            case '\u2016': case '\u2212': case '\u00a2':
            case '\u00a3': case '\u00ac':
            	c = '-'; break;
            }

            descArray[i] = c;
        }
        return new String(descArray);
    }

    // UړI̕R[hɂĂ߂āAΉĂȂ '?' ɂ
    static String narrowCharset(String str, String charsetName) {
        String narrowedDesc = str;
        if (!charsetName.equals("")) {
            try {
                narrowedDesc =
                        new String(str.getBytes(charsetName), charsetName);
                // characters which are not supported by the charset
                // become '?'
            } catch (UnsupportedEncodingException e) {
                Logger.warning(e.toString());
            }

        }
        return narrowedDesc;
    }

    public String getId() {
        return cacheId;
    }

    /**
     * [nl] ID("low"܂܂Ȃ)Ԃ
     * @return ID
     * @since NicoCache_nl+111111mod
     */
    public String getVideoId() {
        return stripLow(cacheId);
    }

    /**
     * [nl] GRm~[LbVH
     * @return GRm~[LbVȂ true
     * @since NicoCache_nl+111111mod
     */
    public boolean isLow() {
    	return isLowSmid(cacheId);
    }

    public File getCacheFile() {
//    	return cacheFile;
    	// [nl] r䂪KvȂ̂ŏid2File擾
    	File cacheFile = id2File_get(cacheId);
    	if (cacheFile != null) {
    		setDescribe(cacheFile, true);
    		return cacheFile;
    	} else if (cacheFileNew == null) {
    		setCacheFileNew();
    	}
    	return cacheFileNew;
    }

    public String getCacheFileName() {
        return getCacheFile().getName();
    }

    public String getURLString() {
        return getCacheFile().toURI().toString();
    }

    public void setDescribe(String desc) {
//    	cacheFile = getDescribedCacheFile(desc);
    	this.desc = desc;
    	cacheFileNew = null;
    }

    // [nl] LbVt@Cݒ
    private void setDescribe(File cacheFile, boolean setPostfix) {
    	Matcher m = CACHE_FILE_PATTERN.matcher(cacheFile.getName());
    	if (m.matches()) {
    		if(m.group(2) != null) desc = m.group(2);
    		if(setPostfix) postfix = m.group(3);
    		cacheFileNew = null;
    	}
    }
    
    public void setTmpDescribe(String title) throws IOException {
        File newName =
                new File(cacheDir, NLTMP_ + getId() + "_"
                        + getSanitizedDescription(title) + postfix);
        if (getCacheTmpFile().renameTo(newName)) {
            id2Tmp.put(getId(), newName);
        }
    }

    public long length() {
        return getCacheFile().length();
    }

    public long tmpLength() throws IOException {
        return getCacheTmpFile().length();
    }

    public boolean exists() {
        return getCacheFile().exists();
    }

    public void touch() {
    	File cacheFile = getCacheFile();
    	if (cacheFile.exists()) {
    		cacheFile.setLastModified(System.currentTimeMillis());
    		statCache.remove(cacheId);
    	}
    }

    public InputStream getInputStream() throws IOException {
        //cacheFile.setLastModified(System.currentTimeMillis());
        if (Boolean.getBoolean("touchCache")) touch();
        return new FileInputStream(getCacheFile());
    }

    /**
     * ꎞt@C̓̓Xg[𓾂
     * @return ꎞt@C̓̓Xg[
     * @throws IOException
     */
    public InputStream getTmpInputStream() throws IOException {
        File cacheTmpFile = getCacheTmpFile();
        return new FileInputStream(cacheTmpFile);
    }

    /**
     * ꎞt@C̏o̓Xg[𓾂
     * [nl] t@C߂ꍇ1؂l߂ăgC
     * @param append ǋLꍇtrue
     * @return ꎞt@C̏o̓Xg[
     * @throws IOException
     */
    public FileOutputStream getTmpOutputStream(boolean append) throws IOException {
    	String truncated = null;
    	do {
    		try {
    			FileOutputStream out = createTmpOutputStream(append);
    			if (truncated != null) {
    				Logger.warning("title truncated: " + truncated);
    			}
    			return out;
    		} catch (FileNotFoundException e) {
    			String mes = e.getMessage();
    			if (desc == null || mes == null || !mes.contains("too long")) {
    				throw e;
    			}
    			id2Tmp.remove(cacheId);
    			tmpFile = null;
    			
    			// t@Ĉ1؂l߂Ă݂
    			truncated = desc.substring(0, desc.length() - 1);
    			setDescribe(truncated);
    			continue;
    		}
    	} while (desc != null && desc.length() > 0);
    	
    	// ܂ŗȂ^CgŕԂ
    	setDescribe(null);
    	return createTmpOutputStream(append);
    }

    // [nl]
    private FileOutputStream createTmpOutputStream(boolean append) throws IOException {
        File cacheTmpFile = getCacheTmpFile();
        return new FileOutputStream(cacheTmpFile, append);
    }

	// LbVK؂ȈʒuEOŊi[B
	// [nl] GRm~[łLΏ悤ɁBU蕪֌Wǉ
	//		GRm~[ł΂̏ꏊɕۑ悤 (#027, #028)
    public void store() throws IOException {
// eco=1̂ŁAID̒ʏGRm~[LbV
// ɍ쐬Ȃ悤r
		synchronized (LockObj.FileStoreLock) {

			//[xmlfix] ꎞt@CꗗXMLLbV폜
			removeXmlCache(TYPE_TEMPLIST, null);

			File cacheTmpFile = getCacheTmpFile();

			File parentDir = cacheDir;
			String storeFolder = "";
			if (!cacheId.endsWith("low")) {
				String lowId = cacheId + "low";
				File lowFile = id2File_get(lowId);
				if (lowFile != null) { 
// low̃LbVt@C^Cg擾
					setDescribe(lowFile, false);
					parentDir = lowFile.getParentFile();
					storeFolder = id2dir.get(lowId);
					Cache.remove(lowId);
				}
// loẅꎞt@C폜
				Cache.removeTmp(lowId);
			} else {
// low̃LbVAʏLbVۑȂ
				String normalId = cacheId.substring(0, cacheId.length()-3);
				File normalCache = id2File_get(normalId);
				if (normalCache != null && normalCache.exists()) {
					cacheTmpFile.delete();
					id2Tmp.remove(cacheId);
					Logger.info("nomal cache already exists, removed: " + cacheId);
					return;
				}
			}

			if (parentDir == cacheDir && Boolean.getBoolean("storeFilter")) {
				storeFolder = suggestStoreFolder();
				if (!storeFolder.equals("")) {
					Logger.info("storing folder: " + storeFolder);
					parentDir = dir2File.get(storeFolder);
				}
			}
			
			setCacheFileNew();
			File cacheFile = new File(parentDir, cacheFileNew.getName());
			
			if (!moveCache(cacheTmpFile, cacheFile)) {
				cacheTmpFile.delete();
			}
			
			id2Tmp.remove(cacheId);
		}
	}

    public boolean deleteTmp() throws IOException {
//        File cacheTmpFile = getCacheTmpFile();
//        cacheTmpFile.delete();
// [nl] {ƃR[hid2TmpɈꎞt@CcĂ܂̂ŏύX
    	File cacheTmpFile = id2Tmp.remove(cacheId);
    	tmpFile = null;
    	return cacheTmpFile != null ? cacheTmpFile.delete() : false;
    }

    boolean existsTmp() { // [nl]
    	return tmpFile != null && tmpFile.exists();
    }

    // [nl] ꎞt@CύX
    // tmpFiletest and setatomicɂ邽߂synchronizedKvB
    protected synchronized File getCacheTmpFile() throws IOException {
        if (tmpFile == null) {
            File knownTmpFile = id2Tmp.get(getId());
            if (knownTmpFile == null) {
                tmpFile = new File(cacheDir, NLTMP_ + getCacheFileName());
            } else {
                tmpFile = knownTmpFile;
            }
        }
        id2Tmp.put(getId(), tmpFile);
        return tmpFile;
    }

}
