package dareka.processor.impl;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.Socket;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.Logger;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.LocalFileResource;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.util.AudioExtractor;
import dareka.processor.util.CommentDownloader;

// [nl] cachetH_֌W̏
public class CacheDirProcessor implements Processor {
    private static final String[] SUPPORTED_METHODS = new String[] { "GET" };
    private static final Pattern CACHE_DIR_PATTERN = Pattern
    	.compile("^http://[^/]+\\.nicovideo\\.jp/cache/?(|.+)$");
    
    public String[] getSupportedMethods() {
		return SUPPORTED_METHODS;
	}

	public Pattern getSupportedURLAsPattern() {
		return CACHE_DIR_PATTERN;
	}

	public String getSupportedURLAsString() {
		return null;
	}

	private static final Pattern TYPE_ID_PATTERN = Pattern.compile(
			"((?:[a-z]{2})?(?:[0-9]+))(?:low)?");
	private static final Pattern LOCAL_FLV_PATTERN = Pattern.compile(
			"^[a-z]{2}[0-9]+(?:low)?\\.(?:flv|swf|mp4)$");
	private static final Pattern AUDIO_EXTRACT_PATTERN = Pattern.compile(
			"^[a-z]{2}[0-9]+(?:low)?\\.(?:mp3|m4a)$");
	private static final Pattern COMMENT_API_PATTERN = Pattern.compile(
			"^_?(\\w{2}\\d+)\\.xml(?:|\\?(.*))$");
	private static final Pattern LST_API_PATTERN = Pattern.compile(
			"^(add|trim)(r)?(list/[^\\?]+)(?:|\\?(.+))$");
	private static final Pattern SEARCH_API_PATTERN = Pattern.compile(
			"^r?search/([^\\?]+)(\\?.+)?$");
	
	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		String path = requestHeader.getPath();
		
		if (path.length() < 7)
			return null;
		path = path.substring(7);

		String smid = "";
		Cache cache = null;
		Matcher mTypeId = TYPE_ID_PATTERN.matcher(path);
		if(mTypeId.find()) {
			smid = mTypeId.group(1);
			cache = new Cache(smid);
			if (!cache.exists()) {
				cache = new Cache(smid + "low");
			}
		}
		
		Matcher m;
		
		// [nl] [JFLViLbV𑗂j
		if (Boolean.getBoolean("localFlv") &&
				(m = LOCAL_FLV_PATTERN.matcher(path)).matches()) {
			if (!cache.exists()) {
				return StringResource.getNotFound();
			}
			Logger.info("Local Flv: " + cache.getId() + " " + cache.getTitle());
			if (Boolean.getBoolean("touchCache")) {
				cache.touch();
			}
			Resource r = new LocalFileResource(cache.getCacheFile());
			LimitFlvSpeedListener.addTo(r);
			r.addCacheControlResponseHeaders(12960000);
			return r;
		}
		
		// t@Co
		if ((m = AUDIO_EXTRACT_PATTERN.matcher(path)).matches()) {
			AudioExtractor ae = new AudioExtractor(
					cache, path, requestHeader.getMessageHeader("User-Agent"));
			return ae.extract();
		}
		
		// Rg擾
		if ((m = COMMENT_API_PATTERN.matcher(path)).matches()) {
			String params = m.group(2);
			if (path.startsWith("_")) {  // e҃Rg(`)
				return CommentDownloader.getResource(smid, "1", requestHeader, browser);
			} else if (params == null) { // p[^(`)
				return CommentDownloader.getResource(smid, "0", requestHeader, browser);
			} else { // p[^LȂV`
				return CommentDownloader.getURLResource(m.group(1),
						params.contains("has_owner_thread=1"), requestHeader);
			}
		}
		
		// K؂ȃ_E[hpXփ_CNg(.01)
		if (path.matches("[a-z]{2}[0-9]+(low)?/(movie|audio)")) {
			if (!cache.exists()) {
				return StringResource.getNotFound();
			}
			String ext = cache.getPostfix();
			String base = "http://www.nicovideo.jp/cache/" + cache.getId();
			if (path.contains("movie")) {
				return StringResource.getRedirect(base + ext);
			} else {
				if (ext.equals(Cache.FLV) || ext.equals(Cache.SWF)) {
					return StringResource.getRedirect(base + ".mp3");
				} else if (ext.equals(Cache.MP4)) {
					return StringResource.getRedirect(base + ".m4a");
				}
			}
			return StringResource.getNotFound();
		}
		// [nl] crossdomain.xmlNotFoundihCȂ̂Ŗ{KvȂj
		else if (path.equals("crossdomain.xml")) {
			return StringResource.getNotFound();
		}
		// [nl] LbVǗy[W
		else if (path.equals("") || path.endsWith(".html")) {
			// [nl] swf̃NGXgȂȈՔłɂ
			if (requestHeader.getMessageHeader("x-flash-version") != null) {
				String flvList = "# my cache\r\n" + Cache.getFlvListForLocalFlv() + "\n";
				return getStringResource(flvList);
			}
			long start = System.currentTimeMillis();
			String str;
			if (path.equals("")) {
				str = Cache.getFlvList();
			} else {
				str = Cache.getFlvList("/local/" + path);
			}
			long elapsed = System.currentTimeMillis() - start;
			Logger.info("Local Flv List (" + elapsed + "msec)");
			return getStringResource(str, "text/html");
		}
		// [nl] [hȂflvidꗗ (FxłLocal Flvp)
		else if (path.equals("flvlist") || path.equals("flvlist/")) {
			String flvList = "# my cache\r\n" + Cache.getFlvListForLocalFlv() + "\n";
			return getStringResource(flvList);
		}
		else if (path.equals("flvlist/crossdomain.xml")) {
			return StringResource.getNotFound();	// "flvlist/"ǉ̂
		}
		// [nl] LbVǗy[W(ajax)
		else if (path.equals("ajax")) {
			long start = System.currentTimeMillis();
			String str = Cache.getFlvListAjax();
			long elapsed = System.currentTimeMillis() - start;
			Logger.info("Local Flv List ajax (" + elapsed + "msec)");
			return getStringResource(str, "application/x-javascript");
		}
		//[nl] + JSON fix
		else if (path.endsWith(".json")) {
			String name = null;
			String str = null;
			long start = System.currentTimeMillis();
			if (path.equals("cachelist.json")) {
				name = "Local Cache List as JSON";
				str = Cache.getCacheListAsJson();
			} else if (path.equals("templist.json")) {
				name = "Local Temp Cache List as JSON";
				str = Cache.getTempListAsJson();
			} else if (path.equals("dirlist.json")) {
				name = "Local Cache Dir List as JSON";
				str = Cache.getDirListAsJson();
			} else if (path.equals("flvlist.json")) {
				name = "Local Flv List as JSON";
				str = Cache.getFlvListAsJson();
			}
			long elapsed = System.currentTimeMillis() - start;

			if (str != null) {
				Logger.info("%s (%d msec)", name, elapsed);
				return getStringResource(str, "application/json");
			}
			return StringResource.getNotFound();
		}
		//[xmlfix] XML擾
		else if (path.startsWith("getxml")) {
			if (path.length() < 7) {
				return StringResource.getBadRequest();
			}
			HashMap<String, String> query = new HashMap<String, String>();
			
			String key, value;
			for (String params : path.substring(7).split("&")) {
				String[] param = params.split("=",2);
				if (params.length() > 1) {
					key = param[0]; value = param[1];
					value = decodeURLForUTF8(value);
					query.put(key, value);
				}
			}
			
			String name = "";
			String xml = null;
			String type = query.get("type");
			
			long start = System.currentTimeMillis();
			// fBNgꗗ
			if (Cache.TYPE_DIRLIST.equals(type)) {
				name = "Local Cache Dir List as XML";
				xml = Cache.getDirListAsXml(query);
			}
			// ꎞt@Cꗗ
			else if (Cache.TYPE_TEMPLIST.equals(type)) {
				name = "Local Temp Cache List as XML";
				xml = Cache.getTempListAsXml(query);
			}
			// LbVꗗ
			else if (Cache.TYPE_CACHELIST.equals(type)) {
				name = "Local Cache List as XML [" + query.get("path") + "]";
				xml = Cache.getCacheListAsXml(query);
			}
			// LbVꗗiׂāj
			else if (Cache.TYPE_CACHELIST_ALL.equals(type)) {
				name = "Local Cache List as XML";
				xml = Cache.getCacheListAsXml(query);
			}
			long elapsed = System.currentTimeMillis() - start;
			
			if (xml != null) {
				//Logger.info("%s (%d msec)", name, elapsed);
				Logger.debug(name + " (" +  elapsed + " msec)");
				StringResource r = new StringResource(xml);
				//r.addNoCacheResponseHeaders();
				r.setResponseHeader("Expires", "Thu, 01 Dec 1994 16:00:00 GMT");
		        r.setResponseHeader("Cache-Control", "no-store");
		        r.setResponseHeader("Pragma", "no-cache");
				r.setResponseHeader("Content-Type", "application/xml; charset=UTF-8");
				return r;
			}
			return StringResource.getBadRequest();
		}
		// O(.??)
		else if (path.equals("log")) {
			if (ViewableLoggerHandler.isLogging()) {
				return new StringResource(ViewableLoggerHandler.getLogWithTableTag());
			} else {
				return new StringResource("NicoCache_nl̃Oy[W猩ɂ́A<br>enableLogHandler=true<br>ƂĂB");
			}
		}
		
		// LbVAPI
		// K\'?'܂łꍇ̂pathfR[hOɏ
		if ((m = SEARCH_API_PATTERN.matcher(path)).matches()) {
			StringBuilder sb = new StringBuilder("{\n");
			String query = URLDecoder.decode(m.group(1), "UTF-8");
			boolean desc = m.group(2) != null && m.group(2).contains("order=d");
			boolean regex = path.startsWith("r");
			for (String s : SearchRewriter.search(query, regex, desc).keySet()) {
				if (Cache.writeJSON(s, sb)) sb.append(",\n");
			}
			if (sb.length() > 2) {
				sb.replace(sb.length() - 2, sb.length(), "\n}");
			} else {
				sb.replace(0, sb.length(), "{}");
			}
			return getStringResource(sb.toString(), "application/json");
		}

		// fR[hAajaxǂ̔
		path = URLDecoder.decode(path, "UTF-8");
		boolean ajax = false;
		if (path.startsWith("ajax_")) {
			ajax = true;
			path = path.substring(5);
		}

		String msg = "NG";
		// [nl] Xgy[W
		// [nl] [JFLVXg폜
		if (path.matches("rm(tmp)?\\?[a-z]{2}[0-9]+(low)?")) {
			smid = mTypeId.group(0);
			if (path.startsWith("rmtmp")) {
				Logger.info("Remove Temporary: " + smid);
				if (Cache.removeTmp(smid))
					msg = "OK";
			} else {
				Logger.info("Remove Cache: " + smid);
// rm?smid (lowtȂ)lowAꎞt@C폜
				if (!smid.endsWith("low")) {
					if (Cache.removeAll(smid)) {
						msg = "OK";
					}
				} else if (Cache.remove(smid)) {
					msg = "OK";
				}
			}
			if (msg.equals("NG")) {
				Logger.info("...failed.");
			}
		}
		// [nl] [JFLVXg^Cgݒ
		else if (path.matches("title\\?[a-z]{2}[0-9]+(low)?(-.*)?")) {
			if (cache.exists()) {
				String title = null;
				String[] t = path.split("-", 2);
				if (t.length == 1) {
					try {	// jRjR擾
						title = new NicoCachingTitleRetriever(smid, requestHeader).call();
					} catch (Exception e) {}
				} else {	// Őݒ
					title = t[1];
				}
				if (title != null) {
					title = Cache.getSanitizedDescription(title);
					if (cache.setTitle(title)) {
						msg = "OK " + title;
					}
				}
			}
		}
		// [nl] [JFLVXgtH_ړ
		else if (path.matches("move\\?([a-z]{2}[0-9]+(low)?)-.*")) {
			if (cache.exists()) {
				String[] t = path.split("-", 2);
				if (cache.moveTo(t[1]))
					msg = "OK";
			}
		}
		// [nl] [JFLVXgtop̎̂݃tH_ړ
		else if (path.matches("topmove\\?([a-z]{2}[0-9]+(low)?)-.*")) {
			if (cache.exists()) {
				String[] t = path.split("-", 2);
				if ("".equals(Cache.getPathFromId(cache.getId())) && cache.moveTo(t[1]))
					msg = "OK";
			}
		}
		// idLbV擾
		else if (path.matches("info\\?[a-z]{2}[0-9]+(low)?")) {
			if (cache.exists()) {
				msg = "OK,"+Cache.getType(smid)+","+cache.getPostfix()+","+cache.length()+",0,"+cache.getTitle();
			} else {
				if (Cache.getDLFinalSize(smid) != 0) {
					msg = "OK,,,"+Cache.getDLSize(smid)+","+Cache.getDLFinalSize(smid)+",";
				} else if (Cache.getDLFinalSize(smid+"low") != 0) {
					msg = "OK,low,,"+Cache.getDLSize(smid+"low")+","+Cache.getDLFinalSize(smid+"low")+",";
				}
			}
		}
		// LSTAPI
		else if ((m = LST_API_PATTERN.matcher(path)).matches()) {
			ajax = true;
			String cmd = m.group(1), file = m.group(3), value = m.group(4);
			boolean regex = m.group(2) != null;
			if (value == null) value = "";
			if ("add".equals(cmd) && value.length() > 0) {
				if (EasyRewriter.LST.append(file, value, regex)) {
					msg = "OK";
				}
			} else if ("trim".equals(cmd)) {
				if (EasyRewriter.LST.trim(file, value.contains("type=smid"))) {
					msg = "OK";
				}
			}
		}
		else {
			Logger.info("no method:" + path);
			return StringResource.getNotFound();
		}
		
		// nhLꍇ͍ŌɉԂ
		if (ajax) {
			return getStringResource(msg);
		} else {
			String redir = null;
			String nlFrom = requestHeader.getParameter("nl-region");
			if (nlFrom != null
					&& (nlFrom.equals("cache") || nlFrom.equals("local")))
				redir = requestHeader.getParameter("nl-from");
			else
				redir = "http://www.nicovideo.jp/";
			StringResource r = StringResource.getRedirect(redir);
			r.addNoCacheResponseHeaders();
			return r;
		}
	}
	
	//[xmlfix]
	private String decodeURLForUTF8(String encodeStr) {
		String decodeStr = (encodeStr == null) ? "" : encodeStr;
		
		try {
			decodeStr = URLDecoder.decode(decodeStr, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			Logger.warning(e.toString());
		}
		
		return decodeStr;
	}
	
	private Resource getStringResource(String str) {
		return getStringResource(str, null);
	}
	
	private Resource getStringResource(String str, String contentType) {
		if (contentType == null) {
			contentType = "text/plain; charset=UTF-8";
		} else if (!contentType.contains("charset=")) {
			contentType += "; charset=UTF-8";
		}
		StringResource r = new StringResource(str);
		r.addCacheControlResponseHeaders(0);
		r.addResponseHeader(HttpHeader.CONTENT_TYPE, contentType);
		return r;
	}
}