package dareka.processor.impl;


import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import java.util.HashMap;
import java.util.concurrent.ConcurrentHashMap;

import dareka.Main;
import dareka.NLConfig;
import dareka.common.CloseUtil;
import dareka.common.Config;
import dareka.common.ConfigObserver;
import dareka.common.FileUtil;
import dareka.common.Logger;
import dareka.common.DiskFreeSpace;
import dareka.common.TextUtil;
import dareka.common.regex.JavaMatcher;
import dareka.common.regex.JavaPattern;
import dareka.extensions.NLFilterListener;
import dareka.extensions.RequestFilter;
import dareka.extensions.Rewriter;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;

class UserFilter {
	static final String REPLACE = new String("[Replace]");
	static final String CONFIG = new String("[Config]");
	static final String REQUEST_HEADER = new String("[RequestHeader]");
	static final String[] SECTIONS = new String[] {
		REPLACE, CONFIG, REQUEST_HEADER
	};
	
	String section, name, id;
	Pattern url;
	FilterPattern match, require, requireHeader, contentType;
	String[] replace;
	boolean multi, each, matchLocal;
	boolean noCache;
	int[] idGroup = { -1, -1 };
	String[] idGroupString;
	String addList, addVariable;
	
	UserFilter() {
		this.section = REPLACE;
	}
	
	UserFilter(String section) {
		for (String constant : SECTIONS) {
			if (constant.equalsIgnoreCase(section)) {
				this.section = constant;
				break;
			}
		}
		if (this.section == null) {
			throw new IllegalArgumentException("invalid section: " + section);
		}
	}
	
	boolean isValid() {
		return name != null && match != null && replace != null &&
				(section != REPLACE || url != null); // [Replace]ȊOURLȗ
	}
}

// IXVł}N($LSTƂ)܂ރp^[ێNX
class FilterPattern {
	UserFilter u;
	boolean not;
	String[] contents;
	JavaPattern[] patterns;
	ArrayList<MacroBase> macros = new ArrayList<MacroBase>();
	
	FilterPattern(UserFilter u, String content, boolean match) {
		this.u = u;
		if (match && u.each) {
			contents = content.split("\0");
		} else {
			contents = new String[1];
			if (content.startsWith("!")) {
				not = true;
				contents[0] = content.substring(1);
			} else {
				contents[0] = content;
			}
		}
		patterns = new JavaPattern[contents.length];
		
		// ̕ӂ͂Ƃc
		if (MacroLST.isMatch(content)) {
			macros.add(new MacroLST(contents));
		}
		if (match && MacroINC.isMatch(content)) {
			macros.add(new MacroINC(contents));
		}
		if (match && MacroSET.isMatch(content)) {
			macros.add(new MacroSET(contents));
		}
		
		for (int i = 0; i < contents.length; i++) {
			updatePattern(i, null);
		}
	}
	
	private boolean updatePattern(int index, FilterContent fc) {
		String content = contents[index];
		for (MacroBase m : macros) {
			content = m.update(content, index);
			if (content == null) {
				patterns[index] = JavaPattern.compile("(?!)");
				return false;
			}
		}
		if (macros.size() > 0) {
			EasyRewriter.debugLog(2, "Update", u.id);
		}
		patterns[index] = JavaPattern.compile(content);
		return true;
	}
	
	/** p^[𓮓IXV */
	boolean update(int index, FilterContent fc) {
		boolean doUpdate = false;
		if (patterns[index] == null) {
			doUpdate = true;
		} else {
			for (MacroBase m : macros) {
				if (doUpdate = m.needsUpdate(index)) {
					break;
				}
			}
		}
		return doUpdate ? updatePattern(index, fc) : true;
	}
	
	/** notlp^[̕v */
	boolean find(String content) {
		assert patterns.length == 1;
		return content != null && update(0, null) &&
				patterns[0].matcher(content).find() ^ not;
	}
	
	// ReplaceɎgeu(nullȂReplaceȂ)
	String replace(String content, int index, FilterContent fc) {
		if (content != null) {
			for (MacroBase m : macros) {
				if ((content = m.found(content, index, fc)) == null) {
					break;
				}
			}
		}
		return content;
	}
	
	/** $name(body)`̃}Np^[ɃRpC */
	static Pattern compile(String name, String body) {
		return Pattern.compile("(?<!\\\\)(?:\\\\\\\\)*\\$" + name +
				"\\(" + body + "\\)");
	}
	
	static Pattern compile(String name) {
		return Pattern.compile("(?<!\\\\)(?:\\\\\\\\)*\\$" + name);
	}
	
	static abstract class MacroBase {
		ArrayList<Object>[] data;  // ėpf[^
		boolean[] changed;         // ύXtO
		
		@SuppressWarnings("unchecked") // ̌^̔z߂ɕKv
		MacroBase(String[] contents) {
			data = new ArrayList[contents.length];
			for (int i = 0; i < data.length; i++) {
				data[i] = new ArrayList<Object>();
			}
			changed = new boolean[contents.length];
		}
		
		// Match̑OupdateĂяoKv邩H
		boolean needsUpdate(int index) { return changed[index]; }
		
		// Match̑OɃp^[蒼߂ɌĂ΂
		// nullԂMatch𒆒f
		// ftHg̓}b`镶폜p^[쐬
		String update(String content, int index/*, FilterContent fc*/) {
			Matcher m = getMatcher(content);
			if (m != null && m.find()) {
				data[index].clear();
				StringBuffer sb = new StringBuffer();
				do {
					updateData(m, index, sb);
				} while (m.find());
				if (sb.length() > 0) {
					content = m.appendTail(sb).toString();
				} else {
					content = m.reset().replaceAll("");
				}
				changed[index] = false;
			}
			return content;
		}
		
		// ftHgŎgpMatcherԂ
		Matcher getMatcher(String str) { return null; }
		
		// ftHgdataXV邽߂ɌĂ΂
		void updateData(Matcher m, int index, StringBuffer sb) {}
		
		// MatcȟɃ}b`xɌĂ΂
		// nullԂReplace𒆒f
		// ftHg͉Ȃ
		String found(String replace, int index, FilterContent fc) {
			return replace;
		}
	}
	
	// $LST
	static class MacroLST extends MacroBase implements EasyRewriter.LST.Observer {
		static Pattern PATTERN = compile(
				"LST", "\"?([^\"\\)]*)\"?");
		
		static boolean isMatch(String content) {
			return PATTERN.matcher(content).find();
		}
		MacroLST(String[] contents) {
			super(contents);
			EasyRewriter.LST.addObserver(this);
		}
		@Override
		boolean needsUpdate(int index) {
			for (Object o : data[index]) {
				if (EasyRewriter.LST.isModified((String) o)) {
					return true;
				}
			}
			return super.needsUpdate(index);
		}
		@Override
		Matcher getMatcher(String content) {
			return PATTERN.matcher(content);
		}
		@Override
		void updateData(Matcher m, int index, StringBuffer sb) {
			String list = m.group(1);
			String group = EasyRewriter.LST.getPattern(list).pattern();
			data[index].add(list);
			m.appendReplacement(sb, Matcher.quoteReplacement(group));
		}
		public void updateLST(String list) {
			for (int i = 0; i < data.length; i++) {
				if (data[i].contains(list)) {
//Logger.debugWithThread("updateLST("+list+")["+i+"]");
					changed[i] = true;
				}
			}
		}
	}
	
	// $INC
	static class MacroINC extends MacroBase {
		static Pattern PATTERN = compile(
				"INC", "(" + FilterContent.NLVAR_CHARS + ")");
		
		static boolean isMatch(String content) {
			return PATTERN.matcher(content).find();
		}
		MacroINC(String[] contents) {
			super(contents);
		}
		@Override
		Matcher getMatcher(String content) {
			return PATTERN.matcher(content);
		}
		@Override
		void updateData(Matcher m, int index, StringBuffer sb) {
			data[index].add(m.group(1));
		}
		@Override
		String found(String replace, int index, FilterContent fc) {
			for (Object o : data[index]) {
				String name = (String) o;
				try {
					int value = Integer.parseInt(fc.getVariable(name, "0"));
					fc.setVariable(name, String.valueOf(++value));
				} catch (NumberFormatException e) {
					Logger.warning("$INC() failed: <nlVar:"+name+"> is not a number.");
				}
			}
			return replace;
		}
	}
	
	// $SET
	static class MacroSET extends MacroBase {
		static Pattern PATTERN = compile(
				"SET", "(" + FilterContent.NLVAR_CHARS + ")\\s*=\\s*([^\\)]*)");
		
		static boolean isMatch(String content) {
			return PATTERN.matcher(content).find();
		}
		MacroSET(String[] contents) {
			super(contents);
		}
		class NameValue {
			String name, value;
			NameValue(String name, String value) {
				this.name = name;
				this.value = value;
			}
		}
		@Override
		Matcher getMatcher(String content) {
			return PATTERN.matcher(content);
		}
		@Override
		void updateData(Matcher m, int index, StringBuffer sb) {
			data[index].add(new NameValue(m.group(1), m.group(2)));
		}
		@Override
		String found(String replace, int index, FilterContent fc) {
			for (Object o : data[index]) {
				NameValue nv = (NameValue) o;
				fc.setVariable(nv.name, nv.value);
			}
			return replace;
		}
	}
	
}

@SuppressWarnings("serial")
class FilterFile extends File {
	public static final String CHARTEST_LINE = "# nltB^`";
	
	private static final ConcurrentHashMap<String, FileUtil.Info> infos =
		new ConcurrentHashMap<String, FileUtil.Info>();
	
	public FilterFile(File parent, String child) {
		super(parent, child);
		putInfoIfNone();
	}
	
	public FilterFile(String pathname) {
		super(pathname);
		putInfoIfNone();
	}
	
	public FilterFile(String parent, String child) {
		super(parent, child);
		putInfoIfNone();
	}
	
	private void putInfoIfNone() {
		if (!infos.containsKey(getPath())) {
			infos.put(getPath(), FileUtil.getInfo(this, null, CHARTEST_LINE));
		}
	}
	
	private FileUtil.Info updateInfo() {
		FileUtil.Info info = FileUtil.getInfo(
				this, infos.get(getPath()), CHARTEST_LINE);
		infos.put(getPath(), info);
		return info;
	}
	
	// ŏɃIuWFNgĂt@CXVĂ邩H
	// canRead,getReader,getWriterĂяo܂ŗL
	boolean isModified() {
		FileUtil.Info info = infos.get(getPath());
		long curr = this.lastModified();
		return info != null ? info.lastModified != curr : 0L != curr;
	}
	
	// ݏIɌĂяo
	void setModified() {
		FileUtil.Info info = infos.get(getPath());
		if (info != null) {
			info.lastModified = this.lastModified();
		}
	}
	
	@Override
	public boolean canRead() {
		updateInfo();
		return this.isFile() && super.canRead();
	}
	
	LineNumberReader getReader() throws IOException {
		InputStreamReader in;
		FileInputStream fis = new FileInputStream(this);
		FileUtil.Info info = updateInfo();
		if (info.charset != null) {
			in = new InputStreamReader(fis, info.charset);
		} else {
			BufferedInputStream bis = new BufferedInputStream(fis);
			info = FileUtil.getInfo(this, info, CHARTEST_LINE, bis);
			in = new InputStreamReader(bis, info.charset);
		}
		return new LineNumberReader(in);
	}
	
	BufferedWriter getWriter(boolean append) throws IOException {
		FileUtil.Info info = updateInfo();
		if (!this.exists()) {
			info.charset = Charset.defaultCharset();
			info.line_separator = System.getProperty("line.separator");
			return new BufferedWriter(
					new OutputStreamWriter(new FileOutputStream(this)));
		} else if (!append) {
			// ǋLȂꍇ͔Ô߃obNAbvcĂ
			this.renameTo(new File(this.getPath().concat(".bak")));
		}
		if (info.charset == null || info.line_separator == null) {
			BufferedInputStream in = new BufferedInputStream(new FileInputStream(this));
			try {
				info = FileUtil.getInfo(this, info, CHARTEST_LINE, in);
			} finally {
				CloseUtil.close(in);
			}
		}
		return new LineSeparatorWriter(new OutputStreamWriter(
				new FileOutputStream(this, append), info.charset), info.line_separator);
	}
	
	class LineSeparatorWriter extends BufferedWriter {
		String line_separator;
		public LineSeparatorWriter(OutputStreamWriter out, String ls) {
			super(out);
			line_separator = ls;
		}
		@Override
		public void newLine() throws IOException {
			if (line_separator == null) {
				super.newLine();
			} else {
				write(line_separator);
			}
		}
	}
	
	int intLength() {
		if (length() > Integer.MAX_VALUE) {
			return Integer.MAX_VALUE;
		}
		return (int)length();
	}
}

// uΏۂ̃RecێNX
class FilterContent implements NLFilterListener.Content {
	/** nlVarɎg镶̏W */
	public static final String NLVAR_CHARS = "[^\\s\"#\\$\\(\\)<>=]+";
	public static final String NLVAR_CONFIG_PREFIX = "config!";
	
	static final Pattern VIDEO_ID_PATTERN = Pattern.compile(
//			"\"videoId\", \"([a-z]{2}\\d+)\"");
// videoId̎擾݂̂Ȃ̂ŃptH[}XlWatchVars͎g킸ɑΉ(Zero)
			"(?<!\\w)videoId(?:\", \"|&quot;:&quot;)([a-z]{2}\\d+)(?!\\w)");
	static final Pattern WATCH_PAGE_PATTERN = Pattern.compile(
			"^http://www\\.nicovideo\\.jp/watch/(\\w{2}\\d+)(\\?.*)?$");
	static final Pattern NLVAR_PATTERN = Pattern.compile(
			"<nlVar:(" + NLVAR_CHARS + ")>");
	
	final String url;
	String content;
	HttpRequestHeader requestHeader;
	HttpResponseHeader responseHeader;
	StringBuffer sb;
	JavaMatcher mc;
	int last, skip;
	
	HashMap<String, String> nlVars = new HashMap<String, String>();
	String id = "", smid = "", memoryId = "", freeSpace = "";
	
	FilterContent(String url, String content,
			HttpRequestHeader requestHeader, HttpResponseHeader responseHeader) {
		this.url = url;
		this.content = content;
		this.requestHeader = requestHeader;
		this.responseHeader = responseHeader;
		sb = new StringBuffer(content.length() * 2);
		
		// watchy[Wsmid擾
		Matcher m = WATCH_PAGE_PATTERN.matcher(url);
		if (m.matches()) {
			smid = m.group(1);
			memoryId = smid;
			if (smid.matches("\\d+")) {
				m = VIDEO_ID_PATTERN.matcher(content);
				if (m.find()) {
					smid = m.group(1);
				}
			}
			if (smid.length() > 2) {
				id = smid.substring(2);
			}
		}
		// LbV󂫗e
		long freeSize = DiskFreeSpace.get(Cache.getCacheDir());
		if (freeSize < Long.MAX_VALUE) {
			freeSpace = String.format("%.2f", freeSize / 1024.0 / 1024 / 1024);
		}
		
		// gݍݕϐ
		nlVars.put("VERSION", Main.VER_STRING);
	}
	
	public String getURL() {
		return url;
	}
	
	public String getContent() {
		return content;
	}
	
	public HttpRequestHeader getRequestHeader() {
		return requestHeader;
	}
	
	public HttpResponseHeader getResponseHeader() {
		return responseHeader;
	}
	
	public String getVariable(String name) {
		return getVariable(name, null);
	}
	
	public String getVariable(String name, String def) {
		String value = null;
		if (name.startsWith(NLVAR_CONFIG_PREFIX)) {
			String config = name.substring(NLVAR_CONFIG_PREFIX.length());
			value = System.getProperty(config);
		}
		if (value == null) {
			value = nlVars.get(name);
		}
		return value != null ? value : def;
	}
	
	public String setVariable(String name, String value) {
		if (value == null) {
			return nlVars.remove(name);
		}
		return nlVars.put(name, value);
	}
	
	String appendVariable(String name, String value, String separator) {
		String oldValue = getVariable(name);
		String newValue = value;
		if (oldValue != null) {
			newValue = oldValue + separator + value;
		}
		nlVars.put(name, newValue);
		return newValue;
	}
	
	String replaceStaticVariables(String s) {
		// ]̂̂͌݊ێ̂nlVarňȂ
		return s.replaceAll("<id>", id)
				.replaceAll("<smid>", smid)
				.replaceAll("<memoryId>", memoryId)
				.replaceAll("<freeSpace>", freeSpace);
	}
	
	String replaceVariables(String s, boolean sys) {
		String value;
		Matcher m = NLVAR_PATTERN.matcher(s);
		if (m.find()) {
			StringBuffer sb = new StringBuffer();
			do {
				if ((value = getReplace(m.group(1), sys)) == null) {
					return null;
				}
				m.appendReplacement(sb, value);
			} while (m.find());
			s = m.appendTail(sb).toString();
		}
		return s;
	}
	
	String getReplace(String name, boolean sys) {
		String value = null;
		if (sys) {
			value = getVariable(name);
			if (value == null) {
				value = EasyRewriter.getExtensionVariable(name, this);
			}
		} else {
			value = EasyRewriter.getExtensionVariable(name, this);
			if (value == null) {
				value = getVariable(name);
			}
		}
		return value;
	}
	
	void setMatcher(JavaPattern pattern) {
		mc = pattern.matcher(content);
		sb.setLength(0);
		last = 0; skip = -1;
	}
	
	boolean find() {
		boolean found = mc.find();
		if (found && mc.start() >= 0 && last >= 0) {
			skip = sb.length() + mc.start() - last;
		} else {
			skip = -1;
		}
		return found;
	}
	
	String group(int num) {
		return mc.group(num);
	}
	
	void appendReplacement(String replacement) {
		mc.appendReplacement(sb, replacement);
		last = mc.end();
	}
	
	String getReplace() {
		return skip < 0 ? null : sb.substring(skip);
		
	}
	
	void updateContent() {
		mc.appendTail(sb);
		content = sb.toString();
		sb.setLength(0);
	}
}

// [nl] XNvgߍ
public class EasyRewriter implements Rewriter, RequestFilter, ConfigObserver {
	
	private static final Pattern NULL_PATTERN = Pattern.compile("(?!)");
	private static final EasyRewriter INSTANCE_SYS = new EasyRewriter();
	private static final EasyRewriter INSTANCE_USR = new EasyRewriter();
	
	private static ArrayList<UserFilter> defaultList = new ArrayList<UserFilter>();
	private static boolean debugMode;
	private static String lastDebugMes;
	private static int lastDebugMesCount;
	private static long checkInterval;
	private static TreeSet<String> idPool = new TreeSet<String>();
	
	static class FilterLists {
		ArrayList<UserFilter> replace = new ArrayList<UserFilter>();
		ArrayList<UserFilter> config  = new ArrayList<UserFilter>();
		ArrayList<UserFilter> header  = new ArrayList<UserFilter>();
	}
	private FilterLists filterLists = new FilterLists();
	private ArrayList<FilterFile> filterFiles;
	private Pattern supportedURLs = NULL_PATTERN;
	private volatile long nextCheckTime;
	
	static {
		Config.addObserver(INSTANCE_SYS);
	}
	
	public void update(Config config) {
		long n = Long.getLong("nlFilterCheckInterval", 1000);
		checkInterval = n > 0 ? n : 0;
	}
	
	// [Config]
	private static HashMap<String, UserFilter> confMap = newConfMap();
	private static HashMap<String, UserFilter> newConfMap() {
		HashMap<String, UserFilter> map = new HashMap<String, UserFilter>();
		for (UserFilter u : INSTANCE_SYS.filterLists.config) {
			map.put(u.name, u);
		}
		for (UserFilter u : INSTANCE_USR.filterLists.config) {
			map.put(u.name, u);
		}
		return map;
	}
	
	// nlFiltergpExtension
	private static ArrayList<NLFilterListener> extensions = new ArrayList<NLFilterListener>();
	
	static boolean addListener(NLFilterListener f) {
		return extensions.add(f);
	}
	
	static String getExtensionVariable(String name, NLFilterListener.Content content) {
		String replaced;
		for (NLFilterListener f : extensions) {
			if ((replaced = f.nlVarReplace(name, content)) != null) {
				return replaced;
			}
		}
		return null;
	}
	
	private EasyRewriter() {} // avoid instantiation.
	
	public static EasyRewriter getInstance_sys() {
		return checkAndLoad(INSTANCE_SYS);
	}
	
	public static EasyRewriter getInstance() {
		return checkAndLoad(INSTANCE_USR);
	}
	
	private static EasyRewriter checkAndLoad(EasyRewriter instance) {
		long currentTime = System.currentTimeMillis();
		if (currentTime > instance.nextCheckTime) {
			// XV`FbNKvƂSXbhUbN
			// bNɑ̃XbhXV`FbNȂ牽Ȃ
			synchronized (instance) {
				if (currentTime > instance.nextCheckTime) {
					instance.load();
					instance.nextCheckTime = currentTime + checkInterval;
				}
			}
		}
		return instance;
	}
	
	static final Pattern IGNORE_FILENAME_PATTERN = Pattern.compile(
			"\\$|^(?:Rs[|Copy)(?: \\(\\d+\\))? (?:`|of) ");
	
	private synchronized void load() {
		ArrayList<FilterFile> newFilterFile = new ArrayList<FilterFile>();
		if (this == INSTANCE_SYS) {
			if (!addFilterFile("nlFilter_sys.txt", newFilterFile)) {
				Logger.warning("no nlFilter_sys.txt found. please check...");
			}
		} else {
			// fBNg猟Aǉ(O)
			File nlFilters = new File("nlFilters");
			if (nlFilters.isDirectory()) {
				String[] filterArray = nlFilters.list(new FilenameFilter() {
					public boolean accept(File dir, String name) {
						if (IGNORE_FILENAME_PATTERN.matcher(name).find() ||
								NLConfig.find("nlFilterIgnore", name)) {
							return false;
						}
						return name.endsWith(".txt");
					}
				});
				if (filterArray != null) {
					Arrays.sort(filterArray);
					for (String fname : filterArray) {
						newFilterFile.add(new FilterFile(nlFilters, fname));
					}
				}
			}
			// nlFilter͍Ōɒǉ
			addFilterFile("nlFilter.txt", newFilterFile);
		}
		
		if (newFilterFile.equals(filterFiles)) {
			boolean modified = false;
			for (FilterFile f : newFilterFile) {
				if (modified = f.isModified()) break;
			}
			if (!modified) {
				if (this == INSTANCE_USR && filterLists.replace.size() == 0) {
					filterLists.replace.addAll(defaultList);
					updateSupportedURLs();
				}
				return;
			}
		}
		
		FilterLists newFilterLists = new FilterLists();
		if (this == INSTANCE_SYS) {
			Logger.info("Loading System Filter:");
		} else {
			Logger.info("Loading User Filters:");
			debugMode = false;
			newFilterLists.replace.addAll(defaultList);
		}
		LST.deleteObservers();
		idPool.clear();
		for (FilterFile f : newFilterFile) {
			Logger.info(f.getPath());
			parseFilterFile(f, newFilterLists);
		}
		filterFiles = newFilterFile;
		filterLists = newFilterLists;
		confMap = newConfMap();
		
		updateSupportedURLs();
	}
	
	// tB^t@C݂Ȃ烊Xgɒǉ
	private boolean addFilterFile(String name, ArrayList<FilterFile> list) {
		FilterFile f = new FilterFile(name);
		if (!f.isFile()) {
			return false;
		}
		return list.add(f);
	}
	
	// tB^ǂݍ
	private void parseFilterFile(FilterFile file, FilterLists lists) {
		LineNumberReader reader = null;
		String readed = null;
		try {
			reader = file.getReader();
			int state = -1;
			UserFilter u = null;
			StringBuilder sb = new StringBuilder(256);
			while ((readed = reader.readLine()) != null) {
				try {
					if (state == -1) {
						if ((u = parseSection(readed)) != null)
							state = 0;
					} else if (state == 0) {
						if (readed.startsWith("#")) { // RgAEg
							continue;
						} else if (readed.equalsIgnoreCase("Match<")) {
							sb.setLength(0);
							state = 1;
						} else if (readed.equalsIgnoreCase("Replace<")) {
							sb.setLength(0);
							state = 2;
						} else if (!parseOption(u, readed)) {
							errorLog("Syntax", file.getPath(), reader, readed);
							state = -1;
						}
					} else {
						if (readed.equals(">")) {
							switch (state) {
							case 1:
								parseMatch(u, sb.toString());
								state = 0;
								break;
							case 2:
								 // Ō̉s폜
								if (!u.each && sb.length() > 0) {
									sb.delete(sb.length() - 2, sb.length());
								}
								parseReplace(u, sb.toString());
								state = -1;
								u.id = newIdentifier(file, u.name);
								addFilter(u, lists);
								break;
							}
						}
						if (u.each && sb.length() > 0) {
							sb.append("\0");
						}
						sb.append(readed);
						// sǉ(y[W̉sR[h͖)
						if (!u.each && state == 2) {
							sb.append("\r\n");
						}
					}
				} catch (PatternSyntaxException e) {
					Logger.error(e);
					state = -1;
				}
			}
		} catch (IOException e) {
			Logger.error(e);
		} finally {
			CloseUtil.close(reader);
		}
	}
	
	// StB^ňӂID𐶐
	private String newIdentifier(File file, String name) {
		if (name == null) {
			return null;
		}
		String base = file.getPath().replaceAll("\\\\", "/")
				.replaceAll("^nlFilters/", "").replaceAll("\\.txt$", "");
		String id = base + "/" + name;
		int i = 2;
		while (!idPool.add(id)) {
			id =  base + "/" + name + "#" + i++;
		}
		return id;
	}
	
	// G[O\
	static void errorLog(
			String kind, String path, LineNumberReader reader, String readed) {
		StringBuilder sb = new StringBuilder(256);
		sb.append(kind).append(" error");
		if (reader != null) {
			sb.append(" at line ").append(reader.getLineNumber());
		}
		if (path != null) {
			sb.append(" in ").append(path);
		}
		if (readed != null) {
			for (String s : readed.split("[\r\n\0]+")) {
				sb.append("\n => ").append(s);
			}
		}
		Logger.warning(sb.toString());
	}
	
	// fobOO\
	static void debugLog(int indent, String mes, String name) {
		if (!debugMode) return;
		
		StringBuilder sb = new StringBuilder(256);
		if (mes != null) {
			for (int i = 0; i < indent; i++) {
				sb.append("  ");
			}
			sb.append("[Debug]").append(mes);
			if (name != null) {
				sb.append(": ").append(name);
			}
		}
		String debugMes = sb.toString();
		if (lastDebugMes == null) {
			lastDebugMes = debugMes;
		} else if (debugMes.equals(lastDebugMes)) {
			lastDebugMesCount++;
		} else {
			sb.setLength(0);
			sb.append(lastDebugMes);
			if (lastDebugMesCount > 0) {
				sb.append(" *").append(++lastDebugMesCount);
			}
			Logger.info(sb.toString());
			lastDebugMes = debugMes.length() > 0 ? debugMes : null;
			lastDebugMesCount = 0;
		}
	}
	
	// ZNVJnp[XUserFilter𐶐
	private UserFilter parseSection(String line) {
		if (line.equalsIgnoreCase("[Debug]")) {
			debugMode = true;
		} else {
			for (String s : UserFilter.SECTIONS) {
				if (line.equalsIgnoreCase(s)) {
					return new UserFilter(s);
				}
			}
		}
		return null;
	}
	
	// IvVp[X
	private static boolean parseOption(UserFilter u, String line) {
		String param[] = line.trim().split("\\s*=\\s*", 2);
		if (param.length < 2) {
			return false;
		} else if (param[0].equalsIgnoreCase("Multi")) {
			u.multi = param[1].equalsIgnoreCase("TRUE");
		} else if (param[0].equalsIgnoreCase("EachLine")) {
			u.each = param[1].equalsIgnoreCase("TRUE");
		} else if (param[0].equalsIgnoreCase("URL")) {
			u.url = Pattern.compile("http://" + param[1]);
		} else if (param[0].equalsIgnoreCase("Name")) {
			u.name = param[1];
		} else if (param[0].equalsIgnoreCase("Require")) {
			u.require = new FilterPattern(u, param[1], false);
		} else if (param[0].equalsIgnoreCase("idGroup")) {
			u.idGroupString = param[1].split("\\s*,\\s*");
			if (u.idGroupString[0].startsWith("!")) {
				u.idGroupString[0] = u.idGroupString[0].substring(1);
				u.noCache = true;
			}
			if (!u.idGroupString[0].matches("\\d+")) {
				return false; // ŏ̃p[^͐̂݋e
			}
			u.idGroup[0] = Integer.parseInt(u.idGroupString[0]);
			if (u.idGroupString.length < 2) {
				u.idGroup[1] = u.idGroup[0];
			} else if (u.idGroupString[1].matches("\\d+")) {
				u.idGroup[1] = Integer.parseInt(u.idGroupString[1]);
			}
		} else if (param[0].equalsIgnoreCase("RequireHeader")) {
			u.requireHeader = new FilterPattern(u, param[1], false);
		} else if (param[0].equalsIgnoreCase("ContentType")) {
			u.contentType = new FilterPattern(u, param[1], false);
		} else if (param[0].equalsIgnoreCase("MatchLocal")) {
			u.matchLocal = param[1].equalsIgnoreCase("TRUE");
		} else if (param[0].equalsIgnoreCase("AddList")) {
			u.addList = param[1];
		} else if (param[0].equalsIgnoreCase("AddVariable")) {
			u.addVariable = param[1];
		} else {
			return false;
		}
		return true;
	}
	
	// Matchp[X
	private static void parseMatch(UserFilter u, String matchStr) {
		u.match = new FilterPattern(u, matchStr, true);
	}
	
	// Replacep[X
	private static void parseReplace(UserFilter u, String replace) {
		replace = replace.replaceAll("<CRLF>", "\r\n").replaceAll("<TAB>", "\t");
		if (u.each) {
			u.replace = replace.split("\0");
		} else {
			u.replace = new String[1];
			u.replace[0] = replace;
		}
	}
	
	// ǂݍ񂾃tB^̑ΏURLւPattern쐬
	private void updateSupportedURLs() {
		// ܂AdPatternO
		TreeSet<String> patterns = new TreeSet<String>();
		for (UserFilter u : filterLists.replace) {
			patterns.add(u.url.pattern());
		}
		// SĂURLORŌqȃp^[쐬
		StringBuilder sb = new StringBuilder();
		for (String s : patterns) {
			// onMatchURLoߑS}b`(.*)Kv
			sb.append("(?:").append(s).append(".*)|");
		}
		if (sb.length() > 0) {
			supportedURLs = Pattern.compile(sb.substring(0, sb.length() - 1));
		} else {
			supportedURLs = NULL_PATTERN;
		}
	}
	
// NX̎擾p
	public static JavaPattern[] getMatch(String name) {
		UserFilter u = confMap.get(name);
		if (u != null) {
			return u.match.patterns;
		}
		return null;
	}
    
	public static String[] getReplace(String name) {
		UserFilter u = confMap.get(name);
		if (u != null) {
			return u.replace;
		}
		return null;
	}
	
	public static Pattern getURL(String name) {
		UserFilter u = confMap.get(name);
		if (u != null) {
			return u.url;
		}
		return null;
	}
	
	public static JavaPattern getRequire(String name) {
		UserFilter u = confMap.get(name);
		if (u != null && u.require != null) {
			return u.require.patterns[0];
		}
		return null;
	}
	
	public static boolean getNotRequire(String name) {
		UserFilter u = confMap.get(name);
		if (u != null && u.require != null) {
			return u.require.not;
		}
		return false;
	}
	
	static boolean checkDebugMode() {
		return debugMode;
	}
	
	private void addFilter(UserFilter u, FilterLists lists) {
		if (u.isValid()) {
			if (u.section == UserFilter.CONFIG) {
				lists.config.add(u);
				Logger.info("  config: " + u.name);
			} else if (u.section == UserFilter.REQUEST_HEADER) {
				if (u.match.patterns.length == u.replace.length) {
					lists.header.add(u);
					Logger.info("  header: " + u.name);
				} else {
					Logger.warning("  ******** headerFilter needs same " +
							"Match&Replace number at " + u.name + " ********");
				}
			} else {
				lists.replace.add(u);
				Logger.info("  " + u.name);
			}
			Logger.debug(u.id);
		} else {
			if (u.name == null) {
				Logger.warning("  ******** Filter has some errors ********");
			} else {
				Logger.warning("  ******** " + u.name + " has some errors ********");
			}
		}
	}
	
	// [hɉe󂯂ȂtB^(.01)
	public static void addSystemFilter(String name, String uri, String match, String replace) {
		UserFilter u = new UserFilter();
		try {
			u.name = name;
			u.url = Pattern.compile(uri);
			parseMatch(u, match);
			parseReplace(u, replace);
			defaultList.add(u);
		} catch(PatternSyntaxException e) {
			Logger.warning(name + "tB^̐K\łB");
		}
	}
	
	// Rewriter interface
	public Pattern getRewriterSupportedURLAsPattern() {
		if (!Main.isDirectoryWatching()) {
			checkAndLoad(this);
		}
		return supportedURLs;
	}
	
	public String onMatch(Matcher match, HttpResponseHeader responseHeader, String content)
			throws IOException {
		String url = match.group();
		return applyUserFilter(url, content, null, responseHeader,
				getMatchedUserFilter(url, null, responseHeader));
	}
	
	/**
	 * ɑ΂ nlFilter Kp
	 * 
	 * @param url tB^Ώۂ URL
	 * @param content tB^Kp镶
	 * @see #replace(String, String, HttpRequestHeader, HttpResponseHeader)
	 */
	public static String replace(String url, String content) {
		return replace(url, content, null, null);
	}
	
	/**
	 * ɑ΂ nlFilter Kp
	 * 
	 * @param url tB^Ώۂ URL (requestHeader w肵ꍇ null )
	 * @param content tB^Kp镶
	 * @param requestHeader NGXgwb_Asvȏꍇ null
	 * @param responseHeader X|Xwb_Asvȏꍇ null
	 * @return tB^
	 * @since NicoCache_nl+120609mod
	 */
	public static String replace(String url, String content,
			HttpRequestHeader requestHeader, HttpResponseHeader responseHeader) {
		if (url == null && requestHeader != null) {
			url = requestHeader.getURI();
		}
		ArrayList<UserFilter> userFilters;
		String replaced;
		userFilters = INSTANCE_SYS.getMatchedUserFilter(
				url, requestHeader, responseHeader);
		replaced = INSTANCE_SYS.applyUserFilter(
				url, content, requestHeader, responseHeader, userFilters);
		userFilters = INSTANCE_USR.getMatchedUserFilter(
				url, requestHeader, responseHeader);
		replaced = INSTANCE_USR.applyUserFilter(
				url, content, requestHeader, responseHeader, userFilters);
		return replaced;
	}
	
	static final Pattern HOST_PATTERN = Pattern.compile(
			"http://([^/]++)/?");
	
	// idGroupɑΉuL̊mFpȂ̂œeɈӖ͖
	static final String[] DUMMY_REPLACES = { "0", "1", "2" };
	
	// RequestFilter interface
	public int onRequest(HttpRequestHeader requestHeader) throws IOException {
		if (!Main.isDirectoryWatching()) {
			checkAndLoad(this);
		}
		String uri = requestHeader.getURI();
		boolean replaced = false;
		for (UserFilter u : filterLists.header) {
			for (int i = 0; i < u.match.patterns.length; i++) {
				if (!u.match.update(i, null)) {
					continue;
				}
				JavaMatcher jm = u.match.patterns[i].matcher(uri);
				if (!jm.matches() || selectReplace(u, uri, jm, DUMMY_REPLACES) == null) {
					continue;
				}
				if (debugMode) {
					Logger.info("[Debug] HeaderFilter : " + u.name + " : " + uri);
				}
				try {
					uri = jm.replaceFirst(u.replace[i]);
				} catch (IndexOutOfBoundsException e) {
					Logger.error(e);
					Logger.warning("invalid replace: " +
							u.replace[i] + " on " + u.name + " at " + i);
					// G[Nꍇ͈Ŝ߃NGXgj
					return RequestFilter.DROP;
				}
				Logger.debugWithThread(" => " + uri + " replaced");
				replaced = true;
			}
		}
		if (replaced) {
			requestHeader.setParameter("nl-URI", requestHeader.getURI());
			requestHeader.setURI(uri);
			Matcher m = HOST_PATTERN.matcher(uri);
			if (m.lookingAt()) {
				requestHeader.setMessageHeader("Host", m.group(1));
			}
		}
		return RequestFilter.OK;
	}
	
	/** URLɃ}b`郆[U[tB^̃Xg𐶐ĕԂ */
	ArrayList<UserFilter> getMatchedUserFilter(String url,
			HttpRequestHeader requestHeader, HttpResponseHeader responseHeader) {
		if (url == null && requestHeader != null) {
			url = requestHeader.getURI();
		}
		ArrayList<UserFilter> matched =
			new ArrayList<UserFilter>(filterLists.replace.size());
		for (UserFilter u : filterLists.replace) {
			if (!u.url.matcher(url).lookingAt()) {
				continue;
			}
			if (u.requireHeader != null && (requestHeader == null ||
					!u.requireHeader.find(requestHeader.toString()))) {
				continue;
			}
			if (u.contentType != null && (responseHeader == null ||
					!u.contentType.find(responseHeader.getMessageHeader(
							HttpHeader.CONTENT_TYPE)))) {
				continue;
			}
			if (LocalDirProcessor.isSupportedURL(url)) {
				// localȊO}b`ꍇ͌݊ێ̂MatchLocalmF
				// localŃ}b`ꍇ͐VKtB^Ȃ̂őfʂ
				int pos = url.indexOf("local/");
				if (pos > 0 && u.url.matcher(url.substring(0, pos)).lookingAt()) {
					if (!u.matchLocal) continue;
				}
			}
			matched.add(u);
		}
		return matched;
	}
	
	/** RecɃ[U[tB^Kp */
	String applyUserFilter(String url, String content,
			HttpRequestHeader requestHeader, HttpResponseHeader responseHeader,
			ArrayList<UserFilter> userFilters) {
		if (userFilters == null || userFilters.isEmpty()) {
			return content;
		}
		FilterContent fc = new FilterContent(
				url, content, requestHeader, responseHeader);
		
		if (extensions.size() > 0 && this == INSTANCE_USR) {
			for (NLFilterListener f : extensions) {
				f.nlFilterBegin(fc);
			}
		}
		for (UserFilter u : userFilters) {
			try {
				debugLog(1, "MatchURL", u.id);
				if (u.require != null && !u.require.find(fc.content)) {
					continue;
				}
				for (int i = 0; i < u.match.patterns.length; i++) {
					if (!u.match.update(i, fc)) {
						continue;
					}
					fc.setMatcher(u.match.patterns[i]);
					doReplace(u, i, fc);
				}
			} catch (Exception e) {
				Logger.warning("FILTER PROCESSING ERROR IN " + u.name);
				Logger.error(e);
			}
		}
		debugLog(0, null, null); // fobOOtbV
		
		return fc.content;
	}
	
	// Rec̒us
	private void doReplace(UserFilter u, int index, FilterContent fc) {
		String[] replaces = initReplaces(u, index, fc);
		boolean replaced = false;
		while (fc.find()) {
			String replace = u.match.replace(
					selectReplace(u, fc.url, fc.mc, replaces), index, fc);
			if (replace != null && doReplacement(u, fc, replace)) {
				replaced = true;
				debugLog(2, "MatchReplace", u.id);
			}
			if (!u.multi) break;
		}
		if (replaced) {
			fc.updateContent();
		} else {
			debugLog(3, "noReplace", u.id);
		}
	}
	
	static final Pattern REPLACE_SPLIT_PATTERN = Pattern.compile(
			"([\\s\\S]+?)\\n*<(\\w*)\\$(\\w*)>\\n*([\\s\\S]+)");
	
	private String[] initReplaces(UserFilter u, int index, FilterContent fc) {
		String[] replaces = new String[3];
		if (index < u.replace.length) {
			replaces[0] = fc.replaceStaticVariables(u.replace[index]);
		} else {
			replaces[0] = "";
		}
		if (u.idGroup[0] > 0) {
			Matcher m = REPLACE_SPLIT_PATTERN.matcher(replaces[0]);
			if (m.matches()) {
				if ("".equals(m.group(2)) && "".equals(m.group(3))) {
					replaces[1] = m.group(1);
					replaces[2] = m.group(4);
				} else {
					replaces[1] = m.group(1) + m.group(2) + m.group(4);
					replaces[2] = m.group(1) + m.group(3) + m.group(4);
				}
			} else {
				replaces[1] = replaces[2] = replaces[0];
			}
		}
		for (int i = 0; i < replaces.length; i++) {
			if (replaces[i] == null || replaces[i].length() == 0) {
				continue;
			}
			// <$>ŊmFKv
			if (replaces[i].contains("$URL")) {
				replaces[i] = replaceURLGroup(u.url.matcher(fc.url), replaces[i]);
			}
		}
		return replaces;
	}
	
	// idGroup̎wɉĒuIĕԂ
	private static String selectReplace(
			UserFilter u, String url, JavaMatcher mc, String[] replaces) {
		if (u.idGroup[0] <= 0) {
			return replaces[0];
		}
		// LbV̑ݏŒutB^
		String eachSmid = null, eachId = null;
		try {
			eachSmid = mc.group(u.idGroup[0]);
			if (u.idGroup[1] > 0) {
				eachId = mc.group(u.idGroup[1]);
			}
		} catch (IndexOutOfBoundsException e) {
			Logger.error(e);
			Logger.warning("invalid idGroup: " + u.name);
			return null;
		}
		if ((eachSmid == null || eachSmid.matches("\\d+")) && eachId != null) {
			eachSmid = Cache.id2Smid(eachId);
		}
		if (eachSmid == null) {
			return null;
		}
		if (TextUtil.isVideoId(eachSmid)) {
			String type = Cache.getType(eachSmid);
			if (!u.noCache && type != null) {
				// LbV݂ꍇɒu
				return replaces[type.equals("low") ? 2 : 1]
						.replaceAll("<eachSmid>", eachSmid);
			} else if (u.noCache && type == null) {
				// LbV݂Ȃꍇɒu
				return replaces[0];
			}
		} else if (extensions.size() > 0) {
			if (u.idGroupString.length == 1) {
				eachId = null;
			} else {
				eachId = u.idGroupString[1];
				if (eachId.matches("URL\\d+")) { // URLO[vQ
					int g = Integer.parseInt(eachId.substring(3));
					Matcher m = u.url.matcher(url);
					if (m.find() && m.groupCount() <= g) {
						eachId = m.group(g);
					}
				}
			}
			for (NLFilterListener f : extensions) {
				int type = f.idGroup(eachSmid, eachId);
				if (type == NLFilterListener.NOT_SUPPORTED) {
					continue;
				} else if (!u.noCache && type != NLFilterListener.NOT_EXISTS) {
					return replaces[type == NLFilterListener.EXISTS_LOW ? 2 : 1]
							.replaceAll("<eachSmid>", eachSmid);
				} else if (u.noCache && type == NLFilterListener.NOT_EXISTS) {
					return replaces[0];
				}
			}
		}
		return null; // uȂ
	}
	
	private boolean doReplacement(UserFilter u, FilterContent fc, String replace) {
		if ((replace = fc.replaceVariables(replace, this == INSTANCE_SYS)) == null) {
			return false;
		}
		
		fc.appendReplacement(replaceTS(replace));
		boolean replaced = true;
		
		if (u.addList != null) {
			String value = fc.getReplace();
			if (value != null && value.length() > 0) {
				LST.append(u.addList, value, false);
				debugLog(2, "AddList(" + u.addList + ")", value);
			}
			replaced = false;
		}
		if (u.addVariable != null) {
			String value = fc.getReplace();
			if (value != null && value.length() > 0) {
				// TODO Zp[^ݒ
				String newValue = fc.appendVariable(u.addVariable, value, "");
				debugLog(2, "AddVariable(" + u.addVariable + ")", newValue);
			}
			replaced = false;
		}
		
		return replaced;
	}
	
	private static ArrayList<Pattern> macroURLPatternCache =
		new ArrayList<Pattern>();
	
	private static Pattern getMacroURLPattern(int index) {
		if (macroURLPatternCache.size() <= index) {
			synchronized (macroURLPatternCache) {
				while (macroURLPatternCache.size() <= index) {
					macroURLPatternCache.add(null);
				}
			}
		}
		Pattern pattern = macroURLPatternCache.get(index);
		if (pattern == null) {
			synchronized (macroURLPatternCache) {
				if ((pattern = macroURLPatternCache.get(index)) == null) {
					pattern = FilterPattern.compile("URL" + index + "(?!\\d)");
					macroURLPatternCache.set(index, pattern);
				}
			}
		}
		return pattern;
	}
	
	// URL̃O[vQƂu
	private String replaceURLGroup(Matcher mURL, String replace) {
		if (mURL.lookingAt()) {
			for (int i = 0; i <= mURL.groupCount(); i++) {
				replace = getMacroURLPattern(i).matcher(replace).replaceAll(
						mURL.group(i) != null ? mURL.group(i) : "");
			}
		}
		return replace;
	}
	
	static final Pattern MACRO_TS_PATTERN = FilterPattern.compile(
			"TS", "([^?)]*)(\\?[^)]+)?");
	
	// ^CX^vu
	private String replaceTS(String replace) {
		Matcher m = MACRO_TS_PATTERN.matcher(replace);
		if (m.find()) {
			StringBuffer sb = new StringBuffer();
			do {
				String replacement;
				if (m.group(1).length() > 0) {
					// t@C݂ꍇ̂݃p[^t
					long lastModified = new File(m.group(1)).lastModified();
					if (lastModified > 0L) {
						replacement = m.group(1) +
								(m.group(2) != null ? m.group(2) : "?") +
								(lastModified / 1000L);
					} else {
						replacement = m.group(1);
					}
				} else {
					// t@Cw肪ꍇ͌ݎɒu
					replacement = String.valueOf(
							System.currentTimeMillis() / 1000L);
				}
				m.appendReplacement(sb, replacement);
			} while (m.find());
			replace = m.appendTail(sb).toString();
		}
		return replace;
	}

	/**
	 *  nlFilterǗLSTɃANZX邽߂̃[eBeBNX
	 *  @since NicoCache_nl+110110mod
	 */
	public static class LST {
		
		static class Entry {
			static final ConcurrentHashMap<String, Entry> entries =
				new ConcurrentHashMap<String, Entry>();
			FilterFile file;
			Pattern pattern;
			TreeSet<String> set;
			String startline;
			private Entry(String list) {
				if (list.startsWith("!")) {
					file = new FilterFile(list.substring(1));
				} else {
					file = new FilterFile(list);
				}
			}
			static String flip(String list) {
				if (list.startsWith("!")) {
					return list.substring(1);
				} else {
					return "!".concat(list);
				}
			}
			static Entry get(String list) {
				Entry entry = entries.get(list);
				if (entry == null) {
					entries.put(list, entry = new Entry(list));
				}
				if (entry.file.isModified()) {
					entry.pattern = null;
					clear(flip(list));
				}
				return entry;
			}
			static void clear(String list) {
				Entry entry = entries.get(list);
				if (entry != null) {
					entry.pattern = null;
				}
			}
		}
		
		static boolean isQuote(String list) {
			return list.startsWith("!");
		}
		
		static boolean isModified(String list) {
			return Entry.get(list).pattern == null;
		}
		
		// LST̍XVĎ邽߂̊ȈՃIuU[o[
		interface Observer {
			public void updateLST(String list);
		}
		
		static final List<Observer> observers =
			Collections.synchronizedList(new ArrayList<Observer>());
		
		static void addObserver(Observer o) {
			observers.add(o);
		}
		
		static void deleteObserver(Observer o) {
			observers.remove(o);
		}
		
		static void deleteObservers() {
			observers.clear();
		}
		
		static void notifyObservers(String list) {
			String flipList = Entry.flip(list);
			Entry.clear(flipList);
			for (Observer o : observers) {
				o.updateLST(list);
				o.updateLST(flipList);
			}
			for (NLFilterListener f : extensions) {
				f.updateLST(list);
				f.updateLST(flipList);
			}
		}
		
		/**
		 * LSTp^[(LST̓eSĂORŌqp^[)ԂB
		 * 
		 * @param list Xg
		 * @return LSTp^[A݂Ȃꍇ"(?!)"p^[ɂĕԂ
		 */
		public static Pattern getPattern(String list) {
			Entry entry = Entry.get(list);
			if (entry.pattern != null) {
				return entry.pattern;
			}
			synchronized (entry) {
				if (readSet(entry, false)) {
					makePattern(list, entry);
				}
			}
			return entry.pattern != null ? entry.pattern : NULL_PATTERN;
		}
		
		/**
		 * LST̖ɕǉBd͒ǉȂB
		 * LSTɐK\̃^܂܂Ȃꍇ́A
		 * LSTp^[ɊSv邩ǂŏd`FbNsB
		 * ^܂ޏꍇ́At@C當ǂݍݏd`FbNsB
		 * 
		 * @param list Xg
		 * @param input ǉ镶
		 * @param regex K\̃^܂ނH
		 * @return ɒǉłtrue
		 */
		public static boolean append(String list, String input, boolean regex) {
			Entry entry = Entry.get(list);
			synchronized (entry) {
				try { // d`FbN
					if (regex) {
						// ^܂ޏꍇ͖ǂݍŔ
						if (readSet(entry, false) && entry.set.contains(input)) {
							return true;
						}
					} else {
						// ^܂܂Ȃꍇ̓p^[}b`Ŕ
						if (entry.pattern == null) {
							readSet(entry, false);
							makePattern(list, entry);
						}
						if (entry.pattern != null &&
								entry.pattern.matcher(input).matches()) {
							return true;
						}
					}
				} finally {
					entry.set = null;
				}
				entry.file.getParentFile().mkdirs();
				boolean written = false;
				BufferedWriter writer = null;
				try {
					writer = entry.file.getWriter(true);
					writer.write(input);
					writer.newLine();
					written = true;
					return true;
				} catch (IOException e) {
					Logger.debugWithThread(e);
					errorLog("Write", entry.file.getPath(), null, input);
				} finally {
					CloseUtil.close(writer);
					if (written) {
						Logger.debugWithThread("append(" + list + "): " + input);
						entry.file.setModified();
						appendPattern(list, entry, input);
					}
				}
			}
			return false;
		}
		
		/**
		 * LSTɕ񂪊܂܂Ă邩H
		 * 
		 * @param list Xg
		 * @param line mF镶
		 * @return 񂪊Svtrue
		 * @since NicoCache_nl+111111mod
		 */
		public static boolean contains(String list, String line) {
			return getPattern(list).matcher(line).matches();
		}
		
		/**
		 * LST當ɊSvs폜āAXtrimďoB
		 * 
		 * @param list Xg
		 * @param line 폜镶
		 * @return oɐtrue
		 * @see #trim(String, boolean)
		 * @since NicoCache_nl+111111mod
		 */
		public static boolean remove(String list, String line) {
			Entry entry = Entry.get(list);
			synchronized (entry) {
				if (readSet(entry, TextUtil.isVideoId(line))) {
					entry.set.remove(line);
					if (writeSet(entry)) {
						Logger.debugWithThread(list + " removed: " + line);
						return true;
					}
				}
			}
			return false;
		}
		
		/**
		 * LST̏dƑŐ󔒂ȂāAXɎŃ\[gďoB
		 * t@C̃Rgs1sڂ̕R[h蕶đSĎ̂ŒӁB
		 * 
		 * @param list Xg
		 * @param smid trueȂ瓮IDƂ݂Ȃă\[g
		 * @return oɐtrue
		 */
		public static boolean trim(String list, boolean smid) {
			Entry entry = Entry.get(list);
			synchronized (entry) {
				if (readSet(entry, smid)) {
					if (writeSet(entry)) {
						Logger.debugWithThread(list + " trimed");
						return true;
					}
				}
			}
			return false;
		}
		
		// t@CǂݍŃZbg쐬
		static boolean readSet(Entry entry, boolean smid) {
			if (!entry.file.canRead()) {
				return false;
			}
			if (smid) {
				entry.set = new TreeSet<String>(Cache.getSmidComparator(false));
			} else {
				entry.set = new TreeSet<String>();
			}
			LineNumberReader reader = null;
			String line = null;
			try {
				reader = entry.file.getReader();
				while ((line = reader.readLine()) != null) {
					if (reader.getLineNumber() == 1 &&
							line.startsWith(FilterFile.CHARTEST_LINE)) {
						entry.startline = line;
					}
					line = line.trim();
					if (line.length() == 0 || line.startsWith("#")) {
						continue;
					}
					entry.set.add(line); // TreeSetɎ荞ŏdȂ
				}
				return true;
			} catch (IOException e) {
				Logger.debugWithThread(e);
				errorLog("Read", entry.file.getPath(), reader, line);
				entry.set = null;
			} finally {
				CloseUtil.close(reader);
			}
			return false;
		}
		
		// Zbgt@Cɏo
		static boolean writeSet(Entry entry) {
			boolean written = false;
			BufferedWriter writer = null;
			try {
				if (entry.set != null && entry.set.size() > 0) {
					writer = entry.file.getWriter(false);
					if (entry.startline != null) {
						writer.write(entry.startline);
						writer.newLine();
					}
					for (String s : entry.set) {
						writer.write(s);
						writer.newLine();
					}
					written = true;
				}
			} catch (IOException e) {
				Logger.debugWithThread(e);
				errorLog("Write", entry.file.getPath(), null, null);
			} finally {
				CloseUtil.close(writer);
				if (written) {
					entry.file.setModified();
				}
				entry.set = null;
			}
			return written;
		}
		
		// p^[쐬
		static void makePattern(String list, Entry entry) {
			if (entry.set != null && entry.set.size() > 0) {
				StringBuilder sb = new StringBuilder(entry.file.intLength());
				sb.append("(");
				boolean quote = isQuote(list);
				for (String s : entry.set) {
					appendBuffer(sb, s, quote);
				}
				compileBuffer(sb, list, entry);
			}
			entry.set = null;
		}
		
		// p^[ɒǉăp^[č쐬
		static void appendPattern(String list, Entry entry, String input) {
			StringBuilder sb;
			if (entry.pattern != null) {
				sb = new StringBuilder(entry.pattern.pattern());
				sb.replace(sb.length() - 1, sb.length(), "|");
			} else {
				sb = new StringBuilder("(");
			}
			appendBuffer(sb, input, isQuote(list));
			compileBuffer(sb, list, entry);
		}
		
		static void appendBuffer(StringBuilder sb, String input, boolean escape) {
			if (escape) {
				sb.append("\\Q").append(input).append("\\E|");
			} else {
				sb.append(input).append("|");
			}
		}
		
		static void compileBuffer(StringBuilder sb, String list, Entry entry) {
			sb.replace(sb.length() - 1, sb.length(), ")");
			try {
				entry.pattern = Pattern.compile(sb.toString());
			} catch (PatternSyntaxException e) {
				Logger.warning("Pattern syntax error: " + entry.file.getPath());
			}
			notifyObservers(list);
		}
	}
	
}

