package dareka.processor.impl;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.EventListener;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.NLConfig;
import dareka.common.CloseUtil;
import dareka.common.Config;
import dareka.common.ConfigObserver;
import dareka.common.LRUMap;
import dareka.common.Logger;
import dareka.common.TextUtil;
import dareka.extensions.CompleteCache;
import dareka.extensions.NLFilterListener;
import dareka.extensions.SystemEventListener;
import dareka.processor.util.GetThumbInfoUtil;

/**
 * [nl] pbP[WXR[vŋLێNX<br>
 * pbP[WXR[vO͋NX̂݃CX^X擾ł
 * @since NicoCache_nl+110424mod
 */
public class NLShared implements ConfigObserver {
	
	// pbP[WXR[vOŃCX^XeNX̃p^[
	private static final Pattern FRIENDS_CLASS_PATTERN = Pattern.compile(
			"dareka\\.(NLMain|processor.+)");
	
	private NLShared() {
		NLConfig.addObserver(this);
	}
	
	public void update(Config config) {
		vid2cidFile = NLConfig.getFile("vid2cidFile");
		vid2cidLastModified = 0L; // ǂݒ
	}
	
	private static final boolean DEBUG = Boolean.getBoolean("dareka.debug");
	
	/** fobO[h̎͒ʏ탍OƂĂo͂ */
	public void debugWithThread2(String message) {
		if (DEBUG) {
			Logger.info(message);
		}
		Logger.debugWithThread(message);
	}
	
	/** pbP[WXR[vŃCX^X𗘗pꍇ͂g */
	static final NLShared INSTANCE = new NLShared();
	
	/**
	 * {@link #FRIENDS_CLASS_PATTERN} ŋꂽNXɂ̂݃CX^XԂB
	 * Ăяo`FbNɃRXĝŁAKvȏꍇ͕ێĎg񂷂ƁB
	 * 
	 * @return NLShared̃CX^X
	 * @throws RuntimeException ĂȂNX̌Ăяo
	 */
	public static NLShared getInstance() {
		String callerName = "(no name)";
		StackTraceElement[] ste = new Throwable().getStackTrace();
		if (ste.length > 1) {
			callerName = ste[1].getClassName();
		}
		if (!FRIENDS_CLASS_PATTERN.matcher(callerName).matches()) {
			throw new RuntimeException("you are not friends: " + callerName);
		}
		return INSTANCE;
	}
	
	
	/**
	 * jRsmidƃNGXgԂێ
	 * @since NicoCache_nl (9).03
	 */
	private static ConcurrentHashMap<String, Long> liveIdList =
		new ConcurrentHashMap<String, Long>();
	
	public void setLiveMovie(String smid) {
		liveIdList.put(smid, new Long(System.currentTimeMillis()));
	}
	
	public boolean isLiveMovie(String smid) {
		// Â͍̂폜
		long now = System.currentTimeMillis();
		for (String oldId : liveIdList.keySet()) {
			Long setTime = liveIdList.get(oldId);
			// Ƃ肠30bɂĂ݂
			if (setTime != null && now > setTime.longValue() + 30000L) {
				liveIdList.remove(oldId);
			}
		}
		return liveIdList.containsKey(smid);
	}
	
	
	/**
	 * LbVrō폜ׂ̃Xg
	 * @since NicoCache_nl (9).10
	 */
	private static Set<String> deleteSet =
		Collections.synchronizedSet(new HashSet<String>());
	
	public boolean isInDeleteSet(String id) {
		return deleteSet.contains(id);
	}
	
	public void addDeleteSet(String id) {
		deleteSet.add(id);
	}
	
	public void removeDeleteSet(String id) {
		deleteSet.remove(id);
	}
	
	
	/**
	 * LbVɌĂяoExtension CacheManager肪K؁H
	 * @since NicoCache_nl (9).10
	 */
	private static List<CompleteCachePair> completeCaches =
		Collections.synchronizedList(new ArrayList<CompleteCachePair>());
	
	class CompleteCachePair {
		public int priority;
		public CompleteCache completeCache;
		CompleteCachePair(int p, CompleteCache c) {
			this.priority = p;
			this.completeCache = c;
		}
	}
	
	public void addCompleteCache(CompleteCache c) {
		completeCaches.add(new CompleteCachePair(0, c));
	}
	
	public ArrayList<CompleteCache> getCompleteEntries() {
		// XgԂOupdateĂяoă\[g
		for (CompleteCachePair p : completeCaches) {
			p.completeCache.update();
			p.priority = p.completeCache.getPriority();
			if (p.priority > 0) p.priority = 0; // dl0ōlȂ̂
		}
		
		// Rs[̂\[gĂ邩A}`Xbh΍͓ɗvȂ͂
		Object[] objArray = completeCaches.toArray();
		Arrays.sort(objArray, new Comparator<Object>() {
			public int compare(Object o1, Object o2) {
				return ((CompleteCachePair) o1).priority - ((CompleteCachePair) o2).priority;
			}
		});
		
		ArrayList<CompleteCache> list = new ArrayList<CompleteCache>();
		for (Object o : objArray) {
			list.add(((CompleteCachePair) o).completeCache);
		}
		return list;
	}
	
	
	/**
	 *  XbhIDsmidւ̑Ήێ
	 *  @since NicoCache_nl+110424mod
	 */
	private static Map<String, String> thread2smidCache =
		Collections.synchronizedMap(new LRUMap<String, String>());
	
	/** XbhɊ֘Atꂽsmid擾A֘AnullԂ */
	public String thread2smid(String thread) {
		if (!TextUtil.isThreadId(thread)) {
			return null;
		}
		if (!thread2smidCache.containsKey(thread)) {
			// LbVɖthumbinfo擾Ă݂
			try {
				GetThumbInfoUtil.get(thread);
			} catch (IOException e) {
				Logger.debugWithThread(e);
			}
		}
		return thread2smidCache.get(thread);
	}
	
	/** Xbhsmid֘At */
	public void putThread2Smid(String thread, String smid) {
		if (TextUtil.isThreadId(thread) && TextUtil.isVideoId(smid)) {
			thread2smidCache.put(thread, smid);
			Logger.debugWithThread("thread2smid: " + thread + " => " + smid);
		}
	}
	
	
	/**
	 * IID(l=vid)ۂɃLbVID(l=cid)ւ̑Ήێ
	 * @since NicoCache_nl+110424mod
	 */
	private static Map<String, String> vid2cid = newTreeMap();
	private static File vid2cidFile;
	private static long vid2cidLastModified;
	
	private static final Pattern VID2CID_LINE_PATTERN = Pattern.compile(
			"\\s*\"(\\d+)\"\\s*:\\s*\"(\\d+)\",?");
	
	private static TreeMap<String, String> newTreeMap() {
		return new TreeMap<String, String>(new Comparator<String>() {
			public int compare(String s1, String s2) {
				// Integer͈̔͂Ɏ܂Ȃ(10ȏ)͂̂
				return (int)(Long.parseLong(s1) - Long.parseLong(s2));
			}
		});
	}
	
	/**
	 * vid Ɋ֘Atꂽ cid 擾A vid ̂܂ܕԂ
	 * @param vid IID(l)
	 * @return ۂɃLbVID(l)
	 */
	public String vid2cid(String vid) {
		loadJson();
		String cid = vid2cid.get(vid);
		if (cid != null) {
			Logger.debugWithThread(
					"vid2cid(get): " + vid + " => " + cid);
			return cid;
		}
		return vid;
	}
	
	/**
	 * vid  cid ֘At(ʂ vid2cidFile ɕۑ)
	 */
	public void put_vid2cid(String vid, String cid) {
		if (!validateVID(vid)) {
			Logger.warning("invalid vid: " + vid);
			return;
		}
		debugWithThread2("vid2cid(put): " + vid + " => " + cid);
		
		synchronized (vid2cid) {
			String prevId = vid2cid.put(vid, cid);
			if (prevId == null || !prevId.equals(cid)) {
				saveJson();
			}
		}
	}
	
	/** videoIdƂėLȕ񂩁H */
	public boolean validateVID(String vid) {
		try {
			// ID9܂ŗL
			if (Integer.parseInt(vid) < 1000000000) {
				return true;
			}
		} catch (NumberFormatException e) {
			Logger.debugWithThread(e);
		}
		return false;
	}
	
	// TODO t@C^ă}bvԂ悤
	private void loadJson() {
		if (vid2cidFile == null || !vid2cidFile.isFile() ||
				vid2cidFile.lastModified() <= vid2cidLastModified) {
			return;
		}
		debugWithThread2("loading " + vid2cidFile.getPath());
		BufferedReader br = null;
		try {
			TreeMap<String, String> map = newTreeMap();
			br = new BufferedReader(new FileReader(vid2cidFile));
			String line; Matcher m;
			while ((line = br.readLine()) != null) {
				if ((m = VID2CID_LINE_PATTERN.matcher(line)).lookingAt() &&
						!NLConfig.matches("deletedVideoId", m.group(1))) {
					map.put(m.group(1), m.group(2));
				}
			}
			vid2cidLastModified = vid2cidFile.lastModified();
			vid2cid = map;
		} catch (IOException e) {
			Logger.error(e);
		} finally {
			CloseUtil.close(br);
		}
	}
	
	// TODO t@Cƃ}bv^悤
	private void saveJson() {
		if (vid2cidFile == null) {
			return;
		}
		debugWithThread2("saving " + vid2cidFile.getPath());
		if (vid2cidFile.exists()) {
			File bakFile = new File(vid2cidFile.getPath() + ".bak");
			bakFile.delete();
			vid2cidFile.renameTo(bakFile);
		}
		BufferedWriter bw = null;
		boolean written = false;
		try {
			bw = new BufferedWriter(new FileWriter(vid2cidFile));
			for (Map.Entry<String, String> e : vid2cid.entrySet()) {
				if (written) {
					bw.append(",\r\n");
				} else {
					bw.append("{\r\n");
					written = true;
				}
				bw.write("\"");
				bw.write(e.getKey());
				bw.write("\": \"");
				bw.write(e.getValue());
				bw.write("\"");
			}
			bw.write("\r\n}");
		} catch (IOException e) {
			Logger.error(e);
		} finally {
			CloseUtil.close(bw);
			if (written) {
				vid2cidLastModified = vid2cidFile.lastModified();
			}
		}
	}
	
	
	/**
	 * CxgʒmC^[tF[X
	 * @since NicoCache_nl+111130mod
	 */
	private static List<SystemEventListener> systemEventListeners =
		Collections.synchronizedList(new ArrayList<SystemEventListener>());
	
	public void addEventListener(EventListener listener) {
		boolean added = false;
		if (listener instanceof SystemEventListener) {
			Logger.debug("adding SystemEventListener: " + listener);
			added |= systemEventListeners.add((SystemEventListener) listener);
		}
		if (listener instanceof NLFilterListener) {
			Logger.debug("adding NLFilterListener: " + listener);
			added |= EasyRewriter.addListener((NLFilterListener) listener);
		}
		if (!added) {
			Logger.debug("unknown EventListener: " + listener);
		}
	}
	
	public int countSystemEventListeners() {
		return systemEventListeners.size();
	}
	
	/**
	 * VXeCxgŜɒʒm
	 * @param id CxgID
	 * @param source CxgIuWFNg
	 * @param checkResultOK ߂l`FbNērŏ𒆒f邩H
	 * @return Cxg
	 */
	public int notifySystemEvent(int id, NLEventSource source, boolean checkResultOK) {
		for (SystemEventListener listener : systemEventListeners) {
			try {
				int result = listener.onSystemEvent(id, source);
				if (checkResultOK && result != SystemEventListener.RESULT_OK) {
					return result;
				}
			} catch (Throwable e) {
				Logger.error(e);
			}
		}
		return SystemEventListener.RESULT_OK;
	}
	
}
