package dareka.processor.impl;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import dareka.common.Logger;
import dareka.processor.HttpRequestHeader;
import dareka.processor.URLResource;

/**
 * Utility class to treat nicovideo API.
 *
 * For maintenance, all of knowledge for nicovideo API should be placed here.
 *
 */
public class NicoApiUtil {
    private NicoApiUtil() {
        // prevent instantiation
    }

    public static String getThumbURL(String type, String id) {
//        return "http://ext.nicovideo.jp/api/getthumbinfo/" + type + id;
        return EXTAPI_URL + "getthumbinfo/" + type + id;
    }

    public static String getThumbTitle(InputStream thumbResponse) {
        SAXParserFactory f = SAXParserFactory.newInstance();
        try {
            SAXParser p = f.newSAXParser();

            class ThumbTitleHandler extends DefaultHandler {
                public String status = "";

                private boolean inTitle;
                public String title = "";

                /* ( Javadoc)
                 * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
                 */
                @Override
                public void startElement(String uri, String localName,
                        String name, Attributes attributes) {
                    if (name.equals("nicovideo_thumb_response")) {
                        status = attributes.getValue("status");
                    } else if (name.equals("title")) {
                        inTitle = true;
                    }
                }

                /* ( Javadoc)
                 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
                 */
                @Override
                public void endElement(String uri, String localName, String name) {
                    if (name.equals("title")) {
                        inTitle = false;
                    }
                }

                /* ( Javadoc)
                 * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
                 */
                @Override
                public void characters(char[] ch, int start, int length) {
                    if (inTitle) {
                        StringBuilder newTitle = new StringBuilder(title);
                        newTitle.append(ch, start, length);
                        title = newTitle.toString();
                    }
                }
            }

            ThumbTitleHandler h = new ThumbTitleHandler();
            InputSource i = new InputSource(thumbResponse);
            p.parse(i, h);

            if (!"ok".equals(h.status)) {
                return null;
            }

            return h.title;
        } catch (ParserConfigurationException e) {
            Logger.error(e);
        } catch (SAXException e) {
            Logger.error(e);
        } catch (IOException e) {
            Logger.error(e);
        }
        return null;
    }

    /*
     * 폜̉
    <?xml version="1.0" encoding="UTF-8"?>
    <nicovideo_thumb_response status="fail">
    <error>
    <code>DELETED</code>
    <description>deleted</description>
    </error>
    </nicovideo_thumb_response>
     *
     * ̉
    <?xml version="1.0" encoding="UTF-8"?>
    <nicovideo_thumb_response status="ok">
    <thumb>
    <video_id>sm9</video_id>
    <title>VEꑰ -ϔY - bcS[IAzt</title>
    <description>bcS[IAztitR[Xo[Wj</description>
    <thumbnail_url>http://tn-skr.smilevideo.jp/smile?i=9</thumbnail_url>
    <first_retrieve>2007-03-06T00:33:00+09:00</first_retrieve>
    <length>5:20</length>
    <view_counter>3835975</view_counter>
    <comment_num>2615213</comment_num>
    <mylist_counter>49903</mylist_counter>
    <last_res_body>M^[\FG[U M^[\F~_ M^[\FfBA M^[\F} ... </last_res_body>
    <watch_url>http://www.nicovideo.jp/watch/sm9</watch_url>
    <thumb_type>video</thumb_type>
    <tags>
    <tag>Azt</tag>
    <tag>bcS[IAzt</tag>
    <tag></tag>
    <tag>y</tag>
    <tag>̉pY</tag>
    <tag>eiLy[</tag>
    <tag>smŌÂ̓</tag>
    <tag>sm9</tag>
    </tags>
    </thumb>
    </nicovideo_thumb_response>
     */

	// [nl] APIANZXp
	public static final String EXTAPI_URL = "http://ext.nicovideo.jp/api/";
	public static final String FLAPI_URL  = "http://flapi.nicovideo.jp/api/";
	private static final Pattern URL_HOST_PATTERN = Pattern.compile("https?://([^/]+)");
	
	/**
	 * [nl] NG}bvɂĕԂ
	 * 
	 * @param query Ώۂ̃NG
	 * @return NG񂩂琶}bv
	 */
	public static Map<String, String> query2map(String query) {
		HashMap<String, String> map = new HashMap<String, String>();
		if (query != null) {
			for (String param : query.split("&")) {
				String[] kv = param.split("=", 2);
				if (kv.length < 2) {
					map.put(kv[0], null);
				} else {
					try {
						map.put(kv[0], URLDecoder.decode(kv[1], "UTF-8"));
					} catch (UnsupportedEncodingException e) {
						Logger.debug(e);
					}
				}
			}
		}
		return map;
	}
	
	/**
	 * [nl] }bvNGɂĕԂ
	 * 
	 * @param map Ώۂ̃}bv
	 * @return }bv琶NG
	 */
	public static String map2query(Map<String, String> map) {
		StringBuilder sb = new StringBuilder();
		if (map != null) {
			for (Map.Entry<String, String> entry : map.entrySet()) {
				try {
					String key = URLEncoder.encode(entry.getKey(), "UTF-8");
					String value = entry.getValue();
					if(value != null) value = URLEncoder.encode(value, "UTF-8");
					if(sb.length() > 0) sb.append('&');
					sb.append(key);
					if(value != null) sb.append('=').append(value);
				} catch (UnsupportedEncodingException e) {
					Logger.debug(e);
				}
			}
		}
		return sb.toString();
	}
	
	/**
	 * [nl] API ƃp[^}bv path 𐶐
	 * 
	 * @param name API 
	 * @param params p[^}bvAsvȏꍇ null
	 * @return path 
	 */
	public static String makePath(String name, Map<String, String> params) {
		String path = name, query = map2query(params);
		if (query.length() > 0) {
			path += "?" + query;
		}
		return path;
	}
	
	/**
	 * [nl] URL ɃANZX邽߂ URLResource 𐶐ĕԂ
	 * 
	 * @param url ANZXΏۂ URL
	 * @param requestHeader NGXgwb_w肵ꍇ̓X|Xwb_M
	 * @param postString POST AsvȂ null
	 * @return  URLResource
	 * @throws IOException X|Xwb_Mɖ肪
	 */
	public static URLResource getURLResource(String url,
			HttpRequestHeader requestHeader, String postString) throws IOException {
		URLResource r = new URLResource(url);
		r.setFollowRedirects(true);
		if (requestHeader != null) {
			requestHeader.setURI(url);
			Matcher m = URL_HOST_PATTERN.matcher(url);
			if (m.find()) {
				requestHeader.setMessageHeader("Host", m.group(1));
			}
			InputStream in = null;
			if (postString != null) {
				in = new ByteArrayInputStream(postString.getBytes());
				requestHeader.setMethod("POST");
			} else {
				requestHeader.setMethod("GET");
			}
			r.getResponseHeader(in, requestHeader);
		}
		return r;
	}
	
	/**
	 * [nl] URL ɃANZXēe𕶎ɂĕԂ
	 * 
	 * @param resource ANZXΏۂ URLResourceAX|Xwb_MĂKv
	 * @return ΎMeAs null
	 * @see #getURLResource(String, HttpRequestHeader, String)
	 */
	public static String getResourceContent(URLResource resource) {
		try {
			byte[] body = resource.getResponseBody();
			if (body != null) return new String(body);
		} catch (IOException e) {
			Logger.debug(e);
		}
		return null;
	}
	
	/**
	 * [nl] flapi ɃANZXēeԂ
	 * 
	 * @param path "/api/`" ȍ~ path 
	 * @param requestHeader NGXgwb_
	 * @param postString POST AsvȂ null
	 * @return ΎMeAs null
	 */
	public static String getFlapiContent(String path,
			HttpRequestHeader requestHeader, String postString) {
		try {
			return getResourceContent(getURLResource(
					FLAPI_URL.concat(path), requestHeader, postString));
		} catch (IOException e) {
			Logger.debug(e);
		}
		return null;
	}
	
	/**
	 * [nl] flapi ɃANZXēeԂ
	 * 
	 * @param name API 
	 * @param params NG̃}bv
	 * @param requestHeader NGXgwb_
	 * @param postString POST AsvȂ null
	 * @return ΎMeAs null
	 * @see #query2map(String)
	 */
	public static String getFlapiContent(String name, Map<String, String> params,
			HttpRequestHeader requestHeader, String postString) {
		return getFlapiContent(makePath(name, params), requestHeader, postString);
	}
	
}
