package dareka.processor.impl;

import java.io.File;
import java.io.IOException;
import java.net.URLDecoder;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import dareka.common.Logger;
import dareka.common.regex.JavaMatcher;
import dareka.common.regex.JavaPattern;
import dareka.extensions.Rewriter;
import dareka.processor.HttpResponseHeader;

public class SearchRewriter implements Rewriter {
	// Rewriter interface
	static final Pattern SUPPORTED_PATTERN = Pattern
			.compile("^http://(?:www|nine)\\.nicovideo\\.jp/(?:search|tag)/([-0-9A-Za-z%+\\.]+)(\\?.*)?$");

	static final Pattern TAG_SEARCH_PATTERN = Pattern
			.compile("http://(?:www|nine)\\.nicovideo\\.jp/tag/");

	public Pattern getRewriterSupportedURLAsPattern() {
		return SUPPORTED_PATTERN;
	}
	
	private String[] writeItem(String smid, String filename) {
		boolean isLow = smid.endsWith("low");
		String noLowID = isLow ? smid.substring(0, smid.length() - 3) : smid;
		String title = "nicocache-unknown-title";
		if (filename.length() > smid.length() + 5) {
			title = filename.substring(smid.length() + 1, filename.length() - 4);
		}
		String background = isLow ? "#fff7f7" : "#f7ffff";
		Cache cache = new Cache(smid);
		String extension = cache.getPostfix();
		
		// yop̃N
		String mp3 = "";
		if (extension.equals(Cache.FLV) || 
			(extension.equals(Cache.MP4) /* && new File("MP4Box.exe").exists() */) ||
			(extension.equals(Cache.SWF) /* && new File("swfextract.exe").exists() */)
			) {
			mp3 = "<a href=\"/cache/" + smid + "/audio\"></a>E";
		}

		String thumbBlock =
			"<p class=\"font12\" style=\"font-weight:bold;\">" + noLowID + "</p>\r\n" +
			"<a href=\"watch/" + noLowID + "\"><img alt=\"" + title + "\" " +
			"src=\"http://tn-skr.smilevideo.jp/smile?i=" + noLowID.substring(2) +
			"\" width=\"130\" height=\"100\" style=\"border:solid 2px #333\"></a>\r\n";
		String descBlock =
			"<p class=\"font12\"><span>"+
			"<a class=\"cacheResultTitle\" href=\"watch/" + noLowID + "\">" + title + "</a></span><br>\r\n" +
			"<span>LbV " + extension +" ...</span></p>\r\n" +
			"<span></span><p class=\"font12\">\n" +
				"<a href=\"/cache/" + smid + "/movie\">ۑ</a>E" +
				mp3 + 
				"<a href=\"/cache/" + noLowID + ".xml\">R</a>E" +
				"<a href=\"/cache/rm?" + smid + "\" onclick=\"return confirm('܂H');\">폜</a>\r\n" +
			"</p>\r\n";

		String[] ret = { thumbBlock, descBlock, background };
		return ret;
	}

	// Y郊NGXg̍ۂɌĂ΂
	public String onMatch(Matcher match, HttpResponseHeader responseHeader, String content)  
			throws IOException {
		if (!Boolean.getBoolean("useSearchExtension") ||
			 Boolean.getBoolean("disableSearchRewriter") ||
				(TAG_SEARCH_PATTERN.matcher(match.group(0)).lookingAt() &&
						!Boolean.getBoolean("insertSearchResultToTagPage"))) {
			return content;
		}
		// 1y[WڂȂ\
		if (match.group(2) != null && match.group(2).contains("page=")) {
			return content;
		}
		
		String query = Cache.tidyTitle(URLDecoder.decode(match.group(1), "UTF-8"));
		Map<String, File> matched = search(query, false, true);
		if (matched.size() == 0) {
			return content;
		}
		
// ʖߍݗpp^[
		JavaPattern pattern;
		String blockString;
		int index = -1;
		while (true) {
			String confName = "SearchExtConf";
			if (index >= 0)
				confName += index;
			index++;
			JavaPattern[] jp = EasyRewriter.getMatch(confName);
			String[] r = EasyRewriter.getReplace(confName);
			JavaPattern require = EasyRewriter.getRequire(confName);
			boolean notRequire = EasyRewriter.getNotRequire(confName);
			if (jp != null) {
				pattern = jp[0]; 
				blockString = r[0];
				if (require == null || (require.matcher(content).find() ^ notRequire)) {
					break;
				}

			} else {
				Logger.info("no SearcExtConf match. do nothing...");
				return content;
			}
		}
		
		StringBuilder sb = new StringBuilder();
		int resultMax = Integer.getInteger("searchResultMax", Integer.MAX_VALUE);
		int resultCount = 0;
		String buf = blockString;
		for (Map.Entry<String, File> e : matched.entrySet()) {
			if (resultCount >= resultMax) {
				break;
			}
			String[] item = writeItem(e.getKey(), e.getValue().getName());
			assert item != null && item.length == 3;
			buf = buf.replaceFirst("<thumbBlock>", Matcher.quoteReplacement(item[0]))
					 .replaceFirst("<descBlock>" , Matcher.quoteReplacement(item[1]))
					 .replaceFirst("<colorBlock>", Matcher.quoteReplacement(item[2]));
			if (!buf.contains("<thumbBlock>")) {
				// <thumbBlock>Ȃ珑oăZbg
				sb.append(buf); buf = blockString;
			}
			resultCount++;
		}
		if (resultCount == 0) {
			return content;
		} else if (buf != blockString) {
			sb.append(buf); // c肪Ώo
		}
		// ߍ
		JavaMatcher jm = pattern.matcher(content);
		if (jm.find()) {
			String resultMes = resultCount + "";
			if (resultCount >= resultMax) {
				resultMes = "ʂ܂B擪" + resultMax + "\܂";
			}
			content = jm.replaceFirst("$0" + Matcher.quoteReplacement(
				"<div><a href=\"javascript:void(0)\" onclick=\"var ele=document.getElementsByName('cacheResult');for(var i=0; i<ele.length; i++){ele[i].style.display=(ele[i].style.display!='' ? '' : 'none');}\">" +
				"<b>NicoCachẽLbV̌  i" + resultMes + "j</b></a></div>" + sb));
		}
		return content;
	}
	
	/**
	 * LbVă}b`ʂid2File}bv𐶐ĕԂB
	 * ʏ팟̏ꍇ́A݊ێׂ̈ɃtH_܂߂Ȃt@CΏۂ
	 * PP(󔒕؂E啶)ORsB
	 * K\̏ꍇ́AtH_܂߂pXΏۂɕvsB
	 * 
	 * @param query 
	 * @param regex K\̏ꍇtrueAʏ팟̏ꍇfalse
	 * @param desc }bv̊i[~ɂꍇtrueAȂfalse
	 * @return }b`LbṼ\[gςid2File}bv
	 * @since NicoCache_nl+110110mod
	 */
	public static SortedMap<String, File> search(String query, boolean regex, boolean desc) {
		SortedMap<String, File> matched = new TreeMap<String, File>(
				Cache.getSmidComparator(desc));
		Map<String, File> caches = Cache.getId2File();
		if (caches.size() == 0) {
			return matched;
		}
		if ("low".equals(query)) { // low ŌƂ̂ݓꏈ
			for (Map.Entry<String, File> e : caches.entrySet()) {
				if (e.getKey().endsWith("low")) {
					matched.put(e.getKey(), e.getValue());
				}
			}
		} else if (regex) {
			try {
				Pattern queryPattern = Pattern.compile(query);
				for (Map.Entry<String, File> e : caches.entrySet()) {
					if (isMatch(queryPattern, e.getValue())) {
						matched.put(e.getKey(), e.getValue());
					}
				}
			} catch (PatternSyntaxException e) {
				Logger.debugWithThread(e);
				Logger.warning("regex search error: " + query);
			}
		} else {
			String[] queryWords = query.toLowerCase().split("\\s+");
			for (Map.Entry<String, File> e : caches.entrySet()) {
				if (isMatch(queryWords, e.getValue())) {
					matched.put(e.getKey(), e.getValue());
				}
			}
		}
		return matched;
	}
	
	// POR
	private static boolean isMatch(String[] queryWords, File cacheFile) {
		String title = cacheFile.getName();
		String titleLow = Cache.tidyTitle(title.toLowerCase());
		for (String word : queryWords) {
			if (word.startsWith("-")) {
				// NOT
				if (titleLow.contains(word.substring(1)))
					return false;
			} else if (!titleLow.contains(word)) {
				// AND
				return false;
			}
		}
		return true;
	}
	
	// K\
	private static boolean isMatch(Pattern queryPattern, File cacheFile) {
		String path = Cache.getPathFromFile(cacheFile);
		if (path != null) {
			return queryPattern.matcher(path).find();
		}
		return false;
	}
}
