package dareka.processor.impl;

import java.io.FileReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.Main;
import dareka.common.LRUMap;
import dareka.common.Logger;
import dareka.common.TextUtil;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;

/**
 * [nl] watch y[Wp[XăvC[ɐݒ肷ϐ⓮̒l擾NX
 * <ul>
 * <li>h̏ꍇ so.addVariable/var Video 擾
 * <li>Zerȍꍇ watchAPIDataContainer 擾
 * <li>IɃjR HTML5 x[XɂȂꍇŋz
 * </ul>
 * @since NicoCache_nl+120609mod
 */
public class WatchVars {
	
	private static final Pattern WATCH_API_DATA_CONTAINER_PATTERN = Pattern.compile(
			"<div id=\"watchAPIDataContainer\"[^>]*>([\\s\\S]+?)</div>");
	private static final Pattern PLAYLIST_DATA_CONTAINER_PATTERN = Pattern.compile(
			"<div id=\"playlistDataContainer\"[^>]*>([\\s\\S]+?)</div>");
	
	private static final Pattern SO_ADDVARIABLE_PATTERN = Pattern.compile(
			"so\\.addVariable\\(\\s*(\".+?\")\\s*,\\s*\"(.*?)\"\\);");
	private static final Pattern VAR_VIDEO_PATTERN = Pattern.compile(
			"<!--\\s+var Video = (\\{[\\s\\S]+?\\});\\s+--></script>");
	
	private static final Pattern WATCH_ID_PATTERN = Pattern.compile(
			"([a-z]{2})?(\\d+)");
	
	private static final Map<String, WatchVars> cache =
			Collections.synchronizedMap(new LRUMap<String, WatchVars>());
	
	/**
	 * WatchVars IuWFNg擾
	 * <p>
	 * ^ꂽ watch y[W̓e WatchVars IuWFNg𐶐ĕԂB
	 * IDƃ^Cg擾łꍇ {@linkplain NicoIdInfoCache} ɋL^B
	 * XɐIuWFNgLbVB
	 * <p>
	 *  http:// n܂镶A12ȉ̕^ꍇA
	 * ^ꂽɓID擾A̓IDɑΉ WatchVars
	 * IuWFNgLbVɎcĂȂ炻ԂB
	 * <p>
	 * ɍŐV̏擾ꍇ́Awatch y[W̓en WatchVars
	 * IuWFNg𐶐KvLB
	 * ȂAIuWFNg̐ɂ͂x̎ԂKvȂ̂ŁARewriter
	 * ̌ĂяoŎO watch y[WʂĂ鎖łꍇA
	 * LbV𗘗pptH[}XIɗLłB
	 * 
	 * @param content watch y[W̓eA URL or ID or XbhID
	 * @return WatchVars IuWFNgALbVɎcĂȂꍇ̂ null
	 */
	public static WatchVars get(String content) {
		if (content == null) {
			return null;
		}
		if (content.length() <= 12 || content.startsWith("http://")) {
			int pos = content.lastIndexOf('/');
			if (pos > 0) {
				content = content.substring(pos + 1);
			}
			Matcher m = WATCH_ID_PATTERN.matcher(content);
			if (!m.find()) {
				return null;
			}
			String videoId = m.group();
			if (TextUtil.isThreadId(videoId)) {
				videoId = NLShared.INSTANCE.thread2smid(videoId);
				if (videoId == null) {
					return null;
				}
			}
			return cache.get(videoId);
		}
		return new WatchVars(content);
	}
	
	private String content, data;
	private Matcher dataMatcher;
	private boolean zeroWatch;
	private JSON json;
	private JSON.JsonObject flashvars;
	private String videoId, videoTitle;
	private HttpRequestHeader requestHeader;
	
	private WatchVars(String content) {
		this.content = content;
		dataMatcher = WATCH_API_DATA_CONTAINER_PATTERN.matcher(content);
		if (dataMatcher.find()) {
			data = TextUtil.unescapeHTML(dataMatcher.group(1));
		} else {
			data = content;
		}
		if (data.startsWith("{")) {
			zeroWatch = true;
			json = JSON.parse(data);
		} else {
			json = harajuku2json(data);
		}
		if (json != null) {
			JSON.JsonValue v = json.getValue("flashvars");
			if (v instanceof JSON.JsonObject) {
				flashvars = (JSON.JsonObject) v;
			}
			videoId = this.getString("videoId");
			if (videoId != null) {
				videoTitle = this.getString("videoTitle");
				NicoCachingTitleRetriever.putTitleCache(videoId, videoTitle);
				cache.put(videoId, this);
			}
		}
		if (Boolean.getBoolean("debugWatchVars")) {
			Logger.info("WatchVars: JSON " + this.getJSON());
			Logger.info("WatchVars: INFO " + videoId + " " + videoTitle);
			StringBuilder sb = new StringBuilder("WatchVars: TAGS");
			for (String tag : getTags()) sb.append(" ").append(tag);
			Logger.info(sb.toString());
		}
	}
	
	// h watch y[W JSON ɂłグ(watchAPI ̃vCXg`)
	private static JSON harajuku2json(String contents) {
		StringBuilder sb = new StringBuilder("{");
		
		Matcher m = SO_ADDVARIABLE_PATTERN.matcher(contents);
		while (m.find()) {
			if (sb.length() > 1) {
				sb.append(",");
			}
			sb.append(m.group(1)).append(":");
			
			// ɂȂ烁oɂĕƐlgKv
			String s = TextUtil.escapeJSON(decodeURL(m.group(2)));
			sb.append('"').append(s).append('"');
		}
		if ((m = VAR_VIDEO_PATTERN.matcher(contents)).find()) {
			// JavaScript  JSON ɕϊ
			String s = m.group(1)
					.replaceAll("\\s+(\\w+):\\t+", "\"$1\":")
					.replaceAll("'", "\"");
			sb.append(",\"videoDetail\":").append(s);
		}
		
		return JSON.parse(sb.append("}").toString());
	}
	
	private static String decodeURL(String value) {
		try {
			return URLDecoder.decode(value, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			Logger.error(e);
		}
		return "(DECODE ERROR)";
	}
	
	/**
	 * p[X JSON č\zĕԂ
	 * @return č\z JSONAp[XłȂꍇ null
	 */
	public String getJSON() {
		if (json != null) {
			return TextUtil.unescapeJSON(json.toString());
		}
		return null;
	}
	
	/**
	 * ID擾
	 * @return IDA擾łȂꍇ null
	 */
	public String getVideoId() {
		return videoId;
	}
	
	/**
	 * ^Cg擾
	 * @return ^CgA擾łȂꍇ null
	 */
	public String getVideoTitle() {
		return videoTitle;
	}
	
	/**
	 * ^O擾
	 * @return ^ÕXgA擾łȂꍇ͋󃊃Xg
	 */
	public List<String> getTags() {
		ArrayList<String> tags = new ArrayList<String>();
		if (json != null) {
			// ܂ Zero dl tagList 𒲂ׂāA tags 擾
			JSON.JsonValue v = json.getValue("videoDetail", "tagList");
			if (v == null) {
				v = json.getValue("videoDetail", "tags");
			}
			if (v instanceof JSON.JsonArray) {
				for (JSON.JsonValue vv : ((JSON.JsonArray) v).elements) {
					String tag;
					if (vv instanceof JSON.JsonObject) {
						tag = ((JSON.JsonObject) vv).getString("tag");
					} else {
						tag = vv.getString();
					}
					if (tag != null) tags.add(tag);
				}
			}
		}
		return tags;
	}
	
	static HttpRequestHeader  REQUEST_HEADER_RIAPI;
	static HttpResponseHeader RESPONSE_HEADER_HTML;
	static HttpResponseHeader RESPONSE_HEADER_JSON;
	static {
		try {
			REQUEST_HEADER_RIAPI = newRequestHeader(
					"http://riapi.nicovideo.jp/api/");
			RESPONSE_HEADER_HTML = new HttpResponseHeader(
					"HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n");
			RESPONSE_HEADER_JSON = new HttpResponseHeader(
					"HTTP/1.1 200 OK\r\nContent-Type: application/json\r\n\r\n");
		} catch (IOException e) {
			Logger.error(e);
		}
	}
	
	private static HttpRequestHeader newRequestHeader(String url) {
		HttpRequestHeader requestHeader = null;
		try {
			requestHeader = new HttpRequestHeader(
					HttpHeader.GET + " " + url + " HTTP/1.1\r\n\r\n");
			requestHeader.setMessageHeader("User-Agent",
					TextUtil.escapeHTML(Main.VER_STRING));
		} catch (IOException e) {
			Logger.error(e);
		}
		return requestHeader;
	}
	
	/**
	 * RXgN^ɓn watch y[W̓e擾
	 * @return watch y[W̓e
	 */
	public String getContent() {
		return content;
	}
	
	/**
	 * watch y[W JSON ɑ΂ nlFilter Œue擾
	 * <p>
	 * ܂ JSON Ŝɑ΂ nlFilter KpŁAX JSON
	 * f[^̓̒lɑ΂ăGXP[vĂ nlFilter
	 * KpAăGXP[vsB
	 * <p>
	 * ΏۂƂȂ JSON ͈ȉ̒ʂ
	 * <ul>
	 * <li>watchAPI ƂĎ擾 JSON
	 * <li>watch y[W watchAPIDataContainer  JSON
	 * </ul>
	 * <p>
	 * uΏۂ̃f[^͈ȉ̒ʂ
	 * <ul>
	 * <li>videoDetail  description ̒l
	 * </ul>
	 * 
	 * @return JSON u watch y[W̓eA
	 * JSON ܂܂Ȃꍇ {@linkplain #getContent()} Ɠ
	 */
	public String getReplacedContent() {
		if (zeroWatch && videoId != null) {
			if (requestHeader == null) {
				requestHeader = newRequestHeader(
						"http://www.nicovideo.jp/watch/" + videoId);
			}
			replaceJSON("videoDetail", "description");
			
			String replaced = addCacheIcon(json.toString());
			if (content != data) {
				StringBuilder sb = new StringBuilder(content);
				sb.replace(dataMatcher.start(1), dataMatcher.end(1),
						TextUtil.escapeHTML(replaced));
				replaced = replacePlaylistData(sb);
			}
			return replaced;
		}
		return content;
	}
	
	// nlFilter g playlistData u
	private String replacePlaylistData(StringBuilder sb) {
		String replaced = sb.toString();
		Matcher m = PLAYLIST_DATA_CONTAINER_PATTERN.matcher(replaced);
		if (m.find()) {
			String playlistData = TextUtil.unescapeHTML(m.group(1));
			sb.replace(m.start(1), m.end(1),
					TextUtil.escapeHTML(addCacheIcon(playlistData)));
			replaced = sb.toString();
		}
		return replaced;
	}
	
	// nlFilter g JSON ̒lu
	private void replaceJSON(String...args) {
		JSON.JsonValue v = this.getValue(args);
		if (v instanceof JSON.JsonString) {
			String s = v.getString();
			s = EasyRewriter.replace(null, s,
					requestHeader, RESPONSE_HEADER_HTML);
			((JSON.JsonString) v).putString(s);
		}
	}
	
	// nlFilter găLbVt
	private String addCacheIcon(String data) {
		return EasyRewriter.replace(null, data,
				REQUEST_HEADER_RIAPI, RESPONSE_HEADER_JSON);
	}
	
	/**
	 * watch y[W JSON l擾ĕŕԂ
	 * <p>
	 * Zerȍꍇ watchAPIDataContainer  JSON ΏۂƂȂB
	 * h̏ꍇ so.addVariable ̒lɍ\zAX var Video ̒l
	 * videoDetail IuWFNgɃ}bv JSON ΏۂƂȂivCXg`jB
	 * h̏ꍇ͓ŋ[I JSON Ƀ}bvĂ邽߈ȉ̈ႢB
	 * <ul>
	 * <li>Zero ł flashvars IuWFNg̒l JSON IuWFNgɂ
	 * <li>Zero ł̓^Ö tagList Ahł tags/lockedTags ƂȂ
	 * <li>Zero ɑ݂ uploaderInfo  viewerInfo ͖
	 * <li>l^Ul̈ȅꍇƕ̏ꍇ
	 * </ul>
	 * <p>
	 * JSON ͊Kw\Ȃ̂ŁAɂ͎擾Ώۂ̒l肷邽߂
	 * o[Kw\ɕw肷î߉ψƂȂjB
	 * Ⴆ΁AVideoDetail  description ̒l擾ꍇA
	 * {@code getString("VideoDetail", "description")} ƌĂяoB
	 * <p>
	 * w肵ꍇA̒l݂Ȃꍇ flashvars Ď擾B
	 * ̂߁AhɂΉɂ flashvars Œڎw肵ȂǂB
	 * 
	 * @param args 擾Ώۂ̒l肷郁o[
	 * @return 擾Ώۂ̕A݂Ȃ or l̂k̏ꍇ null
	 */
	public String getString(String...args) {
		if (json != null) {
			JSON.JsonValue v = getValue(args);
			if (v instanceof JSON.JsonNull) {
				return null;
			} else if (v != null) {
				return v.getString();
			}
		}
		return null;
	}
	
	private JSON.JsonValue getValue(String...args) {
		JSON.JsonValue v = json.getValue(args);
		if (v == null && args.length == 1 && flashvars != null) {
			return flashvars.members.get(args[0]);
		}
		return v;
	}
	
	/**
	 * watch y[W JSON l擾Đ^UlŕԂ
	 * @param args 擾Ώۂ̒l肷郁o[
	 * @return 擾Ώۂ̒l݂^̏ꍇ̂ true
	 * @see #getString(String...)
	 */
	public boolean getBoolean(String...args) {
		if (json != null) {
			JSON.JsonValue v = getValue(args);
			if (v instanceof JSON.JsonTrue) {
				return true;
			} else if (v != null) {
				return Boolean.getBoolean(v.getString());
			}
		}
		return false;
	}
	
	/**
	 * watch y[W JSON l擾Đl(int)ŕԂ
	 * @param args 擾Ώۂ̒l肷郁o[
	 * @return 擾Ώۂ̐lA݂Ȃ or ls or l̂[̏ꍇ 0
	 * @see #getString(String...)
	 */
	public int getInteger(String...args) {
		if (json != null) {
			JSON.JsonValue v = getValue(args);
			if (v instanceof JSON.JsonNumber) {
				return ((JSON.JsonNumber) v).getInteger();
			} else if (v != null) {
				try {
					return Integer.parseInt(v.getString());
				} catch (NumberFormatException e) {
					Logger.debug(e);
				}
			}
		}
		return 0;
	}
	
	/**
	 * watch y[W JSON l擾Đl(long)ŕԂ
	 * @param args 擾Ώۂ̒l肷郁o[
	 * @return 擾Ώۂ̐lA݂Ȃ or ls or l̂[̏ꍇ 0
	 * @see #getString(String...)
	 */
	public long getLong(String...args) {
		if (json != null) {
			JSON.JsonValue v = getValue(args);
			if (v instanceof JSON.JsonNumber) {
				return ((JSON.JsonNumber) v).getLong();
			} else if (v != null) {
				try {
					return Long.parseLong(v.getString());
				} catch (NumberFormatException e) {
					Logger.debug(e);
				}
			}
		}
		return 0;
	}
	
	/**
	 * watch y[W JSON l擾Ďl(double)ŕԂ
	 * @param args 擾Ώۂ̒l肷郁o[
	 * @return 擾Ώۂ̎lA݂Ȃ or ls or l̂[̏ꍇ 0
	 * @see #getString(String...)
	 */
	public double getDouble(String...args) {
		if (json != null) {
			JSON.JsonValue v = getValue(args);
			if (v instanceof JSON.JsonNumber) {
				return ((JSON.JsonNumber) v).getDouble();
			} else if (v != null) {
				try {
					return Double.parseDouble(v.getString());
				} catch (NumberFormatException e) {
					Logger.debug(e);
				}
			}
		}
		return 0;
	}
	
	// eXgp
	public static void main(String args[]) {
		for (int i = 0; i < args.length; i++) {
			FileReader fr = null;
			StringBuilder sb = new StringBuilder();
			try {
				fr = new FileReader(args[i]);
				int c;
				while ((c = fr.read()) != -1) {
					sb.append((char) c);
				}
				fr.close();
				String input = TextUtil.unescapeHTML(sb.toString());
				System.out.println("INPUT = " + input);
				JSON json = JSON.parse(input);
				if (json != null) {
					System.out.println("PARSE = " + json);
				}
			} catch (IOException e) {
				System.out.println(e.toString());
			}
		}
	}
	
}

// KɎ JSON p[T
// See: http://www.json.org/json-ja.html
class JSON {
	static abstract class JsonValue {
		String value;
		public JsonValue(String value) {
			this.value = value;
		}
		public String toString() {  // encoded
			return value;
		}
		public String getString() { // decoded
			return value;
		}
	}
	static class JsonObject extends JsonValue {
		Map<String, JsonValue> members = new LinkedHashMap<String, JsonValue>();
		public JsonObject(String value) {
			super(value);
		}
		public String toString() {
			StringBuilder sb = new StringBuilder("{");
			for (Map.Entry<String, JsonValue> e : members.entrySet()) {
				if (sb.length() > 1) {
					sb.append(",");
				}
				sb.append('"').append(e.getKey()).append('"')
					.append(":").append(e.getValue().toString());
			}
			return sb.append("}").toString();
		}
		public String getString() {
			return TextUtil.unescapeJSON(this.toString());
		}
		public String getString(String member) {
			JsonValue v = members.get(member);
			return v != null ? v.getString() : null;
		}
	}
	static class JsonArray extends JsonValue {
		List<JsonValue> elements = new ArrayList<JsonValue>();
		public JsonArray(String value) {
			super(value);
		}
		public String toString() {
			StringBuilder sb = new StringBuilder("[");
			for (JsonValue v : elements) {
				if (sb.length() > 1) {
					sb.append(",");
				}
				sb.append(v.toString());
			}
			return sb.append("]").toString();
		}
		public String getString() {
			return TextUtil.unescapeJSON(this.toString());
		}
	}
	static class JsonString extends JsonValue {
		public JsonString(String value) {
			super(value);
		}
		public String toString() {
			return "\"" + value + "\"";
		}
		public String getString() {
			return TextUtil.unescapeJSON(value);
		}
		public String putString(String replace) {
			String oldValue = value;
			value = TextUtil.escapeJSON(replace);
			return oldValue;
		}
	}
	static class JsonNumber extends JsonValue {
		String floor;
		public JsonNumber(String value) {
			super(value);
			int dot = value.indexOf('.');
			floor = dot > 0 ? value.substring(0, dot) : value;
		}
		public int getInteger() {
			return Integer.parseInt(floor);
		}
		public long getLong() {
			return Long.parseLong(floor);
		}
		public double getDouble() {
			return Double.parseDouble(value);
		}
	}
	static class JsonTrue extends JsonValue {
		public JsonTrue(String value) {
			super(value);
		}
	}
	static class JsonFalse extends JsonValue {
		public JsonFalse(String value) {
			super(value);
		}
	}
	static class JsonNull extends JsonValue {
		public JsonNull(String value) {
			super(value);
		}
	}
	
	static JSON parse(String input) {
		if (input != null && input.startsWith("{")) {
			JSON json = new JSON(input);
			if (json.object != null) {
				return json;
			}
		}
		return null;
	}
	
	private String input;
	private int index;
	private JsonObject object;
	private boolean shownDebugDetail;
	
	private JSON(String input) {
		this.input  = input;
		this.index  = -1;
		this.object = (JsonObject) nextValue();
	}
	
	private void debugLog(String message) {
		char c = index < input.length() ? input.charAt(index) : 0;
		if (!shownDebugDetail) {
			Logger.debug("JSON: input = " + input.substring(0, index) +
					c + " <<ERROR at " + index);
			shownDebugDetail = true;
		}
		Logger.debug("JSON: " + message);
	}
	
	public String toString() {
		return object != null ? object.toString() : "{}";
	}
	
	public String getString(String...args) {
		JsonValue v = getValue(args);
		return v != null ? v.getString() : null;
	}
	
	public String putString(String replace, String...args) {
		JsonValue v = getValue(args);
		if (v instanceof JsonString) {
			return ((JsonString) v).putString(replace);
		}
		return null;
	}
	
	public JsonValue getValue(String...args) {
		if (args.length == 1) {
			return object.members.get(args[0]);
		}
		return getValue(object, new ArrayList<String>(Arrays.asList(args)));
	}
	
	private JsonValue getValue(JsonValue v, ArrayList<String> a) {
		if (v != null) {
			if (a.isEmpty()) {
				return v;
			}
			if (v instanceof JsonObject) {
				String s = a.remove(0);
				return getValue(((JsonObject) v).members.get(s), a);
			}
		}
		return null;
	}
	
	private JsonValue nextValue() {
		index++;
		switch (skipBlankChar()) {
		case '{':
			return parseObject();
		case '[':
			return parseArray();
		case '"':
			return parseString();
		}
		return parseLiteral();
	}
	
	private JsonObject parseObject() {
		if (checkStartChar('{')) {
			JsonObject object = new JsonObject("{object}");
			JsonString string;
			JsonValue value;
			for (char c;;) {
				index++;
				if ((string = parseString()) == null || skipBlankChar() != ':' ||
						(value = nextValue()) == null) {
					break;
				}
				object.members.put(string.value, value);
				if ((c = skipBlankChar()) == '}') {
					index++;
					return object;
				} else if (c != ',') {
					break;
				}
			}
		}
		debugLog("object parse error");
		return null;
	}
	
	private JsonArray parseArray() {
		if (checkStartChar('[')) {
			JsonArray array = new JsonArray("[array]");
			JsonValue value;
			for (char c;;) {
				if ((value = nextValue()) == null) {
					break;
				}
				array.elements.add(value);
				if ((c = skipBlankChar()) == ']') {
					index++;
					return array;
				} else if (c != ',') {
					break;
				}
			}
		}
		debugLog("array parse error");
		return null;
	}
	
	private JsonString parseString() {
		if (checkStartChar('"')) {
			int start = index + 1;
			for (char c;;) {
				if ((c = nextChar()) == '"') {
					String value = input.substring(start, index++);
					return new JsonString(value);
				} else if (c == '\\') {
					if ((c = nextChar()) == 'u') {
						index += 4;
					}
					// Ƀp[XȂ \[\"/bfnrt] ȊO̓G[
				} else if (c == 0) {
					break;
				}
			}
		}
		debugLog("string parse error");
		return null;
	}
	
	private JsonValue parseLiteral() {
		char c = skipBlankChar();
		int start = index;
		if (isLower(c)) {
			skipLowerChar();
			int end = index;
			if (isValueEnd(skipBlankChar())) {
				String value = input.substring(start, end);
				if ("null".equals(value)) {
					return new JsonNull(value);
				} else if ("true".equals(value)) {
					return new JsonTrue(value);
				} else if ("false".equals(value)) {
					return new JsonFalse(value);
				}
			}
			debugLog("literal parse error");
		} else if (isDigit(c)) {
			JsonNumber number = parseNumber();
			if (number != null) {
				return number;
			}
			debugLog("number parse error");
		}
		return null;
	}
	
	private JsonNumber parseNumber() {
		char c = skipBlankChar();
		int start = index;
		if (c == '-') {
			c = nextChar();
		}
		if (isDigit(c)) {
			if (c == '0') {
				c = nextChar();
			} else {
				while (isDigit(c = nextChar()))
					;
			}
			if (c == '.') {
				if (!isDigit(c = nextChar())) {
					return null;
				}
				while (isDigit(c = nextChar()))
					;
				if (c == 'e' || c == 'E') {
					c = nextChar();
					if (c == '+' || c == '-') {
						c = nextChar();
					}
					if (!isDigit(c)) {
						return null;
					}
					while (isDigit(c = nextChar()))
						;
				}
			}
			int end = index;
			if (isValueEnd(skipBlankChar())) {
				return new JsonNumber(input.substring(start, end));
			}
		}
		return null;
	}
	
	private boolean isBlank(char c) {
		return c <= 0x20;
	}
	
	private boolean isDigit(char c) {
		return '0' <= c && c <= '9';
	}
	
	private boolean isLower(char c) {
		return 'a' <= c && c <= 'z';
	}
	
	private boolean isValueEnd(char c) {
		return c == ',' || c == '}' || c == ']';
	}
	
	private char nextChar() {
		if (++index < input.length()) {
			return input.charAt(index);
		}
		return 0;
	}
	
	private char skipBlankChar() {
		for (char c; index < input.length(); index++) {
			if (!isBlank(c = input.charAt(index))) {
				return c;
			}
		}
		return 0;
	}
	
	private char skipLowerChar() {
		for (char c; index < input.length(); index++) {
			if (!isLower(c = input.charAt(index))) {
				return c;
			}
		}
		return 0;
	}
	
	private boolean checkStartChar(char start) {
		char c;
		if ((c = skipBlankChar()) == start) {
			return true;
		}
		debugLog("invalid start char '" + c + "'");
		return false;
	}
	
}
